package de.landsh.opendata.csw2dcat;

import de.landsh.opendata.csw2dcat.vocabulary.*;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.jena.datatypes.RDFDatatype;
import org.apache.jena.datatypes.xsd.XSDDatatype;
import org.apache.jena.iri.IRI;
import org.apache.jena.iri.IRIFactory;
import org.apache.jena.rdf.model.*;
import org.apache.jena.vocabulary.DCAT;
import org.apache.jena.vocabulary.DCTerms;
import org.apache.jena.vocabulary.RDF;
import org.apache.jena.vocabulary.VCARD4;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.Node;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.time.LocalDate;
import java.util.*;
import java.util.stream.Collectors;

/**
 * This class performs the GMD -> DCAT-AP.de conversion.
 */
public class MDMetadata2Dataset {

    private static final String LANGUAGE_DE = "de";
    private static final String LANGUAGE_EN = "en";

    private static final Logger log = LoggerFactory.getLogger(MDMetadata2Dataset.class);
    private static final String XPATH_OPERATION_GET_CAPABILITIES = "gmd:identificationInfo/*/srv:containsOperations/srv:SV_OperationMetadata[srv:operationName/gco:CharacterString ='GetCapabilities']";
    static IRIFactory factory = IRIFactory.iriImplementation();
    private final Model model;
    private final Map<String, String> keywordMapping = new HashMap<>();
    private final Map<String, JSONObject> url2license = new HashMap<>();
    Map<String, String[]> themeMapping = new HashMap<>();
    boolean ignoreInvalidMapping = true;
    private CswInterface cswInterface;
    private BridgeSettings settings;

    public MDMetadata2Dataset(Model model, CswInterface cswInterface, BridgeSettings bridgeSettings) {
        this(model);
        this.settings = bridgeSettings;
        this.cswInterface = cswInterface;
    }

    public MDMetadata2Dataset(Model model) {
        this.model = model;
        themeMapping.put("AGRI", new String[]{"AGRI", "farming", "imageryBaseMapsEarthCover", "inlandWaters", "oceans"});
        themeMapping.put("ENVI", new String[]{"ENVI", "biota", "climatologyMeteorologyAtmosphere", "elevation", "environment", "imageryBaseMapsEarthCover", "inlandWaters", "oceans", "geoscientificInformation", "farming", "utilitiesCommunication"});
        themeMapping.put("REGI", new String[]{"REGI", "boundaries", "location", "planningCadastre", "geoscientificInformation", "structure", "imageryBaseMapsEarthCover"});
        themeMapping.put("GOVE", new String[]{"GOVE", "boundaries", "elevation", "imageryBaseMapsEarthCover", "location", "planningCadastre", "utilitiesCommunication"});
        themeMapping.put("TECH", new String[]{"TECH", "climatologyMeteorologyAtmosphere", "elevation", "geoscientificInformation", "imageryBaseMapsEarthCover"});
        themeMapping.put("ECON", new String[]{"ECON", "economy"});
        themeMapping.put("HEAL", new String[]{"HEAL", "health"});
        themeMapping.put("JUST", new String[]{"JUST", "intelligenceMilitary"});
        themeMapping.put("SOCI", new String[]{"SOCI", "society"});
        themeMapping.put("EDUC", new String[]{"EDUC", "society"});
        themeMapping.put("TRAN", new String[]{"TRAN", "transportation", "inlandWaters", "oceans", "structure"});
        themeMapping.put("ENER", new String[]{"ENER", "utilitiesCommunication"});
        themeMapping.put("INTR", new String[]{"INTR"});

        final JSONArray keywordArray = new JSONArray(new JSONTokener(Objects.requireNonNull(getClass().getResourceAsStream("/keyword-mapping.json"))));
        for (int i = 0; i < keywordArray.length(); i++) {
            final JSONObject entry = keywordArray.getJSONObject(i);
            final String key = entry.keys().next();
            final String value = entry.getString(key);
            keywordMapping.put(key, value);
        }

        JSONArray licenses = new JSONArray(new JSONTokener(Objects.requireNonNull(getClass().getResourceAsStream("/licenses.json"))));
        for (Object l : licenses) {
            JSONObject license = (JSONObject) l;
            url2license.put(license.getString("uri"), license);
        }
    }

    static String getTextOrNull(Node node) {
        if (node == null) return null;
        return StringUtils.trimToNull(node.getStringValue());
    }

    // Add a literal to a resource if the specified text if not null
    private static void addLiteralIfNotNull(Resource resource, Property property, String content) {
        if (content != null) {
            resource.addLiteral(property, content);
        }
    }

    /**
     * Convert a <code>gmd:EX_TemporalExtent</code> element to <code>dct:temporal</code> and add it to the datset.
     */
    static void addTemporalExtend(Resource dataset, Element temporalExtend) {

        if (temporalExtend == null) {
            return;
        }

        final Element extend = temporalExtend.element("extent");
        if (extend == null) {
            throw new IllegalArgumentException("gmd:EX_TemporalExtent without gmd:extent child element");
        }

        final String beginPosition;
        final String endPosition;

        final Element childElement = extend.elements().get(0);
        if ("TimePeriod".equals(childElement.getName())) {
            beginPosition = childElement.elementText("beginPosition");
            endPosition = childElement.elementText("endPosition");
        } else if ("TimeInstant".equals(childElement.getName())) {
            beginPosition = childElement.elementText("timePosition");
            endPosition = childElement.elementText("timePosition");
        } else {
            log.warn("unexpected child element of temporal extent: {}", childElement.getName());
            beginPosition = null;
            endPosition = null;
        }

        boolean hasBeginPosition = StringUtils.isNotEmpty(beginPosition);
        boolean hasEndPosition = StringUtils.isNotEmpty(endPosition);

        if (hasBeginPosition || hasEndPosition) {
            final Model model = dataset.getModel();
            final Resource periodOfTime = model.createResource();
            periodOfTime.addProperty(RDF.type, DCTerms.PeriodOfTime);

            if (hasBeginPosition) {
                final RDFDatatype dataType = beginPosition.matches("^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]$") ?
                        XSDDatatype.XSDdate : XSDDatatype.XSDdateTime;
                periodOfTime.addLiteral(SchemaOrg.startDate, model.createTypedLiteral(beginPosition, dataType));
            }
            if (hasEndPosition) {
                final RDFDatatype dataType = endPosition.matches("^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]$") ?
                        XSDDatatype.XSDdate : XSDDatatype.XSDdateTime;
                periodOfTime.addLiteral(SchemaOrg.endDate, model.createTypedLiteral(endPosition, dataType));
            }

            dataset.addProperty(DCTerms.temporal, periodOfTime);
        }

    }

    static boolean isValidURLandIRI(String url) {
        try {
            new URL(url);
        } catch (MalformedURLException ignored) {
            return false;
        }

        final IRI iri = factory.create(url);
        return !iri.hasViolation(false);
    }

    /**
     * Create a FOAF Organization based on the specified CI_ResponsibleParty element.
     */
    static Resource createFoafOrg(Element ciResponsibleParty, Model model) {
        Resource result = model.createResource();
        result.addProperty(RDF.type, FOAF.Organization);

        String name = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:organisationName/gco:CharacterString"));
        String mail = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:electronicMailAddress/gco:CharacterString"));
        String phone = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:phone/gmd:CI_Telephone/gmd:voice/gco:CharacterString"));
        String homepage = repairURL(getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:onlineResource/gmd:CI_OnlineResource/gmd:linkage/gmd:URL")));
        String streetAddress = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:deliveryPoint/gco:CharacterString"));
        String city = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:city/gco:CharacterString"));
        String postalCode = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:postalCode/gco:CharacterString"));
        String region = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:administrativeArea/gco:CharacterString"));
        String country = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:country/gco:CharacterString"));

        addLiteralIfNotNull(result, FOAF.name, name);
        if (mail != null) {
            // email address must be valid
            String uri = "mailto:" + mail;
            if (isValidURLandIRI(uri)) {
                result.addProperty(FOAF.mbox, model.createResource(uri));
            } else {
                log.info("Invalid email address {}", mail);
            }
        }

        if (phone != null)
            try {
                result.addProperty(FOAF.phone, model.createResource(TelephoneNumberClean.numberToURI(phone)));
            } catch (IllegalArgumentException ignore) {
                log.info("Invalid phone number {}", phone);
            }
        if (homepage != null) {
            if (!homepage.startsWith("http")) {
                homepage = "http://" + homepage;
            }
            if (isValidURLandIRI(homepage)) {
                result.addProperty(FOAF.homepage, model.createResource(homepage));
            } else {
                log.info("Invalid homepage {}", mail);
            }
        }

        final Resource address = model.createResource();
        address.addProperty(RDF.type, Locn.Address);
        addLiteralIfNotNull(address, Locn.thoroughfare, streetAddress);
        addLiteralIfNotNull(address, Locn.postName, city);
        addLiteralIfNotNull(address, Locn.postCode, postalCode);
        addLiteralIfNotNull(address, Locn.adminUnitL2, region);
        addLiteralIfNotNull(address, Locn.adminUnitL1, country);

        result.addProperty(Locn.address, address);

        return result;
    }

    static Resource createVcardOrganization(Element ciResponsibleParty, Model model) {
        Resource result = model.createResource();
        result.addProperty(RDF.type, VCARD4.Organization);

        String name = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:organisationName/gco:CharacterString"));

        Element ciAddress = (Element) ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address");
        String mail = getTextOrNull(ciAddress.selectSingleNode("gmd:electronicMailAddress/gco:CharacterString"));
        String streetAddress = getTextOrNull(ciAddress.selectSingleNode("gmd:deliveryPoint/gco:CharacterString"));
        String city = getTextOrNull(ciAddress.selectSingleNode("gmd:city/gco:CharacterString"));
        String postalCode = getTextOrNull(ciAddress.selectSingleNode("gmd:postalCode/gco:CharacterString"));
        String region = getTextOrNull(ciAddress.selectSingleNode("gmd:administrativeArea/gco:CharacterString"));
        String country = getTextOrNull(ciAddress.selectSingleNode("gmd:country/gco:CharacterString"));
        String phone = getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:phone/gmd:CI_Telephone/gmd:voice/gco:CharacterString"));
        String homepage = repairURL(getTextOrNull(ciResponsibleParty.selectSingleNode("gmd:contactInfo/gmd:CI_Contact/gmd:onlineResource/gmd:CI_OnlineResource/gmd:linkage/gmd:URL")));

        addLiteralIfNotNull(result, VCARD4.fn, name);
        if (mail != null) {
            // email address must be valid
            final String uri = "mailto:" + mail;
            if (isValidURLandIRI(uri)) {
                result.addProperty(VCARD4.hasEmail, model.createResource(uri));
            } else {
                log.info("Invalid email address {}", mail);
            }
        }
        if (homepage != null) {
            if (!homepage.startsWith("http")) {
                homepage = "http://" + homepage;
            }
            if (isValidURLandIRI(homepage)) {
                result.addProperty(VCARD4.hasURL, model.createResource(homepage));
            } else {
                log.info("Invalid homepage address {}", homepage);
            }
        }

        if (phone != null) {
            try {
                Resource hasTelephone = model.createResource();
                hasTelephone.addProperty(RDF.type, VCARD4.Voice);
                hasTelephone.addProperty(VCARD4.hasValue, model.createResource(TelephoneNumberClean.numberToURI(phone)));
                result.addProperty(VCARD4.hasTelephone, hasTelephone);
            } catch (IllegalArgumentException ignore) {
                log.info("Invalid phone number {}", phone);
            }
        }

        final Resource hasAddress = model.createResource();
        hasAddress.addProperty(RDF.type, VCARD4.Address);
        addLiteralIfNotNull(hasAddress, VCARD4.street_address, streetAddress);
        addLiteralIfNotNull(hasAddress, VCARD4.locality, city);
        addLiteralIfNotNull(hasAddress, VCARD4.postal_code, postalCode);
        addLiteralIfNotNull(hasAddress, VCARD4.region, region);
        addLiteralIfNotNull(hasAddress, VCARD4.country_name, country);

        result.addProperty(VCARD4.hasAddress, hasAddress);

        return result;
    }

    /**
     * Add a bounding box as dct:spacial property to the dataset based on the specified EX_GeographicBoundingBox element.
     */
    static void addBoundingBox(Resource dataset, Element elementBoundingBox) {
        if (elementBoundingBox == null)
            return;

        if (!"EX_GeographicBoundingBox".equals(elementBoundingBox.getName())) {
            throw new IllegalArgumentException("Element must be a EX_GeographicBoundingBox.");
        }

        String west = getTextOrNull(elementBoundingBox.selectSingleNode("gmd:westBoundLongitude/gco:Decimal"));
        String east = getTextOrNull(elementBoundingBox.selectSingleNode("gmd:eastBoundLongitude/gco:Decimal"));
        String north = getTextOrNull(elementBoundingBox.selectSingleNode("gmd:northBoundLatitude/gco:Decimal"));
        String south = getTextOrNull(elementBoundingBox.selectSingleNode("gmd:southBoundLatitude/gco:Decimal"));

        boolean isLatitudeEqual = NumberUtils.toDouble(north) == NumberUtils.toDouble(south);
        boolean isLongitudeEqual = NumberUtils.toDouble(west) == NumberUtils.toDouble(east);

        final String wkt;
        if (isLongitudeEqual && isLatitudeEqual) {
            // POINT
            wkt = "POINT(" + west + " " + north + ")";
        } else if (isLongitudeEqual || isLatitudeEqual) {
            // LINESTRING
            wkt = "LINESTRING(" + west + " " + south + "," + east + " " + north + ")";
        } else {
            wkt = "POLYGON((" + west + " " + north + "," + east + " " + north + "," + east + " " + south + "," + west + " " + south + "," + west + " " + north + "))";
        }
        final Model model = dataset.getModel();
        final Resource location = model.createResource();
        location.addProperty(RDF.type, DCTerms.Location);
        location.addLiteral(
                Locn.geometry,
                model.createTypedLiteral(wkt, "http://www.opengis.net/ont/geosparql#wktLiteral"));

        dataset.addProperty(DCTerms.spatial, location);
    }

    private static String repairURL(String textOrNull) {
        if (textOrNull == null) {
            return null;
        }

        return textOrNull.replaceAll(" ", "%20");
    }

    /**
     * Collect <code>gmd:CI_ResponsibleParty</code> elements for the specified dataset.
     */
    static Map<String, Element> collectResponsibleParties(Element metadata) {
        final Map<String, Element> responsibleParties = new HashMap<>();

        final List<Node> responsiblePartyList = new ArrayList<>();

        responsiblePartyList.addAll(metadata.selectNodes("gmd:contact/gmd:CI_ResponsibleParty"));
        responsiblePartyList.addAll(metadata.selectNodes("gmd:identificationInfo/*/gmd:pointOfContact/gmd:CI_ResponsibleParty"));

        for (Node responsibleParty : responsiblePartyList) {
            final Node codeListValue = responsibleParty.selectSingleNode("gmd:role/gmd:CI_RoleCode/@codeListValue");
            final String role = codeListValue == null ? null : codeListValue.getText();
            responsibleParties.put(role, (Element) responsibleParty);
        }

        log.debug("Found {} responsible parties.", responsibleParties.size());
        return responsibleParties;
    }

    static void addNamespaces(Element metadata) {
        metadata.addNamespace("gco", "http://www.isotc211.org/2005/gco");
        metadata.addNamespace("gmx", "http://www.isotc211.org/2005/gmx");
        metadata.addNamespace("gmd", "http://www.isotc211.org/2005/gmd");
        metadata.addNamespace("xlink", "http://www.w3.org/1999/xlink");
        metadata.addNamespace("srv", "http://www.isotc211.org/2005/srv");
    }

    static String getServiceEndpoint(Element metadata) {
        String url;
        // Is it a service with a comprehensive description of operations?
        if (metadata.selectSingleNode(XPATH_OPERATION_GET_CAPABILITIES) != null) {
            final Element operationMetadata = (Element) metadata.selectSingleNode(XPATH_OPERATION_GET_CAPABILITIES);
            url = repairURL(StringUtils.trim(operationMetadata.selectSingleNode("srv:connectPoint/gmd:CI_OnlineResource/gmd:linkage/gmd:URL").getStringValue()));
        } else {
            url = repairURL(getTextOrNull(metadata.selectSingleNode("gmd:identificationInfo[1]/*/gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:MD_Identifier/gmd:code")));
            log.warn("Using identifier as service endpoint: {}", url);
        }
        return url;
    }

    public BridgeSettings getSettings() {
        return settings;
    }

    /**
     * Collect date information contained in the gmd:identificationInfo element of the metadata
     */
    Map<String, Element> collectDates(Element metadata) {
        final Map<String, Element> result = new HashMap<>();
        final List<Node> dates = metadata.selectNodes("gmd:identificationInfo/*/gmd:citation/gmd:CI_Citation/gmd:date/gmd:CI_Date");

        for (Node dateElement : dates) {
            final String dateType = getTextOrNull(dateElement.selectSingleNode("gmd:dateType/gmd:CI_DateTypeCode/@codeListValue"));
            final Element date = ((Element) dateElement).element("date");
            final String value = getDateValue(date);
            if (StringUtils.isNotBlank(value)) {
                result.put(dateType, date);
            }
        }

        return result;
    }

    /**
     * Convert the metadata of a service into a DCAT distribution.
     *
     * @param metadata element MD_Metadata of the service
     * @return dcat:Distribution resource object
     */
    Collection<Resource> convertServiceToDistributions(Element metadata) {
        addNamespaces(metadata);

        if (!"MD_Metadata".equals(metadata.getName())) {
            throw new IllegalArgumentException("Input must be a gmd:MD_Metadata element.");
        }
        final String serviceId = getTextOrNull(metadata.selectSingleNode("gmd:fileIdentifier/gco:CharacterString"));

        metadata.addNamespace("gmd", "http://www.isotc211.org/2005/gmd");
        if (null == metadata.selectSingleNode("gmd:hierarchyLevel/gmd:MD_ScopeCode[@codeListValue='service']")) {
            throw new IllegalArgumentException("Input is not a service");
        }

        addNamespaces(metadata);

        int numberOfTransferOptions = metadata.selectNodes("gmd:distributionInfo/*/gmd:transferOptions/*/gmd:onLine/*[gmd:function/*/@codeListValue='download' and gmd:linkage/*[text()]]").size();
        if (numberOfTransferOptions > 1) {
            log.debug("Multi-format service");
            return convertMultiFormatDataset(metadata, true);
        }

        final String url = getServiceEndpoint(metadata);
        if (!isValidURLandIRI(url)) {
            log.warn("Distribution has an invalid URI {} - ignoring it.", url);
            return Collections.emptySet();
        }

        final Resource distribution = model.createResource(url);
        distribution.addProperty(RDF.type, DCAT.Distribution);
        distribution.addProperty(DCAT.accessURL, model.createResource(url));
        // distribution.addProperty(DCAT.downloadURL, model.createResource(url));
        distribution.addProperty(DCAT.accessService, model.createResource(settings.baseIRI + serviceId));

        final String title = getTextOrNull(metadata.selectSingleNode("gmd:identificationInfo/*/gmd:citation/gmd:CI_Citation/gmd:title/gco:CharacterString"));
        if (title != null) {
            distribution.addLiteral(DCTerms.title, model.createLiteral(title.trim(), LANGUAGE_DE));
        }

        final String description = getTextOrNull(metadata.selectSingleNode("gmd:identificationInfo/*/gmd:abstract/gco:CharacterString"));
        if (description != null) {
            distribution.addLiteral(DCTerms.description, model.createLiteral(description.trim(), LANGUAGE_DE));
        }

        final String dateStamp = getTextOrNull(metadata.selectSingleNode("gmd:dateStamp/gco:Date"));
        if (dateStamp != null) {
            distribution.addLiteral(DCTerms.modified, model.createTypedLiteral(dateStamp, XSDDatatype.XSDdate));
        }

        final String format = getTextOrNull(metadata.selectSingleNode("gmd:identificationInfo/*/srv:serviceType/gco:LocalName"));
        if (format != null) {
            distribution.addProperty(DCTerms.format, Mapping.mapFormat(format));
        }

        // license
        final List<Node> otherLegalConstraints = metadata.selectNodes("gmd:identificationInfo/*/gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:otherConstraints/gco:CharacterString");
        addLicenseInformation(distribution, otherLegalConstraints);

        return Collections.singleton(distribution);
    }

    private Collection<Resource> convertMultiFormatDataset(Element metadata, boolean isService) {
        final String serviceId = isService ? getTextOrNull(metadata.selectSingleNode("gmd:fileIdentifier/gco:CharacterString")) : null;
        final List<Node> otherLegalConstraints = metadata.selectNodes("gmd:identificationInfo/*/gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:otherConstraints/gco:CharacterString");
        final String dateStamp = getTextOrNull(metadata.selectSingleNode("gmd:dateStamp/gco:Date"));

        final List<Node> onlineResourceDownload = metadata.selectNodes("gmd:distributionInfo/*/gmd:transferOptions/*/gmd:onLine/*[gmd:function/*/@codeListValue='download' and gmd:linkage/*[text()]]");
        log.debug("Found {} download online resources.", onlineResourceDownload.size());

        final List<Resource> result = new LinkedList<>();
        for (Node transferOption : onlineResourceDownload) {
            final Resource distribution = convertOnlineResourceToDistribution(transferOption, otherLegalConstraints, dateStamp, serviceId);
            if (distribution != null) {
                result.add(distribution);
            }
        }

        return result;
    }

    /**
     * Converts the content of a <code>&lt;gmd:onLine></code> element (e.g., a <code>&lt;gmd:CI_OnlineResource></code>)
     * into a Distribution {@link Resource}.
     */
    private Resource convertOnlineResourceToDistribution(Node onlineResource, List<Node> otherLegalConstraints, String dateStamp, String serviceId) {

        final String url = repairURL(getTextOrNull(onlineResource.selectSingleNode("gmd:linkage/gmd:URL")));
        String name = getTextOrNull(onlineResource.selectSingleNode("gmd:name/*"));
        String description = getTextOrNull(onlineResource.selectSingleNode("gmd:description/*"));

        if (!isValidURLandIRI(url)) {
            log.warn("Distribution has an invalid URI {} - ignoring it.", url);
            return null;
        }

        if ((name == null || StringUtils.equals(name, url)) && StringUtils.isNotEmpty(description)) {
            // There are cases where the URL is repeated in the "name" field and the actual name
            // is in the description field.
            name = description;
            description = null;
        }

        final Resource distribution = model.createResource(url);
        distribution.addProperty(RDF.type, DCAT.Distribution);
        distribution.addProperty(DCAT.accessURL, model.createResource(url));
        distribution.addProperty(DCAT.downloadURL, model.createResource(url));
        if (serviceId != null) {
            distribution.addProperty(DCAT.accessService, model.createResource(settings.baseIRI + serviceId));
        }

        if (StringUtils.isNotEmpty(name)) {
            distribution.addProperty(DCTerms.title, name);
        }
        if (StringUtils.isNotEmpty(description)) {
            distribution.addProperty(DCTerms.description, description);
        }

        if (dateStamp != null) {
            distribution.addLiteral(DCTerms.modified, model.createTypedLiteral(dateStamp, XSDDatatype.XSDdate));
        }

        final String applicationProfile = getTextOrNull(onlineResource.selectSingleNode("gmd:applicationProfile/*"));

        Resource format = null;
        if (StringUtils.startsWith(applicationProfile, "http://publications.europa.eu/resource/authority/file-type/")) {
            format = ResourceFactory.createResource(applicationProfile);
        } else if (StringUtils.isNotEmpty(applicationProfile)) {
            format = Mapping.mapFormat(applicationProfile);
        }

        if (format == null) {
            // Guess the format.
            format = guessFormat(url);
        }

        if (format == null) {
            // some additional heuristics to guess the format
            if ("feed link des dienstes".equals(StringUtils.lowerCase(name))
                    || "feedlink des dienstes".equals(StringUtils.lowerCase(name))
                    || "feed-link des dienstes".equals(StringUtils.lowerCase(name))) {
                format = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/ATOM");
            } else if ("AtomFeed Viewer".equals(name)) {
                format = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/HTML");
            } else if (url.contains("feed") && url.toLowerCase().contains("atom") && url.endsWith(".xml")) {
                format = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/ATOM");
            }

        }


        if (format != null) {
            distribution.addProperty(DCTerms.format, format);
        }

        // license
        addLicenseInformation(distribution, otherLegalConstraints);
        return distribution;
    }

    private void addLicenseInformation(Resource resource, List<Node> otherLegalConstraints) {
        final JSONObject licenseInformation = findLicenseInformation(otherLegalConstraints);

        if (licenseInformation != null) {
            resource.addLiteral(DCATAPde.licenseAttributionByText, licenseInformation.getString("quelle"));

            if (licenseInformation.has("url")) {
                String licenseURL = licenseInformation.getString("url");

                if (licenseURL.startsWith("http://dcat-ap.de/def/licenses/")) {
                    resource.addProperty(DCTerms.license, model.createResource(licenseURL));
                } else {
                    final JSONObject license = url2license.get(licenseURL);

                    if (license != null && license.has("uri")) {
                        resource.addProperty(DCTerms.license, model.createResource(license.getString("uri")));
                    } else {
                        log.debug("Unknown license: {}", licenseURL);
                        handleUnknownLicense(resource, licenseInformation.getString("id"));
                    }
                }
            } else {
                handleUnknownLicense(resource, licenseInformation.getString("id"));
            }
        }
    }

    private void handleUnknownLicense(Resource resource, String licenseId) {
        licenseId = URLEncoder.encode(licenseId, StandardCharsets.UTF_8)
                .replaceAll("%2F", "/"); // Replace the encoded slash with the original slash
        resource.addProperty(DCTerms.license, model.createResource("http://dcat-ap.de/def/licenses/" + licenseId));
        log.debug("Falling back to: {}", licenseId);
    }

    private Resource guessFormat(String url) {
        final String lowerCaseURL = url.toLowerCase();
        if (lowerCaseURL.contains("service=wms")) {
            if (!lowerCaseURL.contains("outputformat=")) {
                return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/WMS_SRVC");
            } else {
                log.info("WMS with output format: {}", url);
                return null;
            }
        } else if (lowerCaseURL.contains("service=wfs") && lowerCaseURL.contains("getcapabilities")) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/WFS_SRVC");
        } else if (lowerCaseURL.contains("service=wfs")) {
            if (!lowerCaseURL.contains("outputformat=")) {
                // assume that the WFS returns GML als standard output format
                return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/GML");
            } else if (lowerCaseURL.contains("outputformat=shape-zip")) {
                return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/SHP");
            } else if (lowerCaseURL.contains("outputformat=shape%2bzip")) {
                return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/SHP");
            } else if (lowerCaseURL.contains("outputformat=csv")) {
                return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/CSV");
            } else {
                log.info("WFS with unknown output format: {}", url);
                return null;
            }
        } else if (lowerCaseURL.endsWith(".html")) {
            return ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/HTML");
        }

        return null;
    }

    /**
     * Convert a {@link Document} containin a gmd:MD_Metadata element into a DCAT-AP.de compliant dcat:Dataset {@link Resource}.
     */
    public Resource convert(Document input) {
        return convert(input.getRootElement());
    }

    /**
     * Convert commons information used for dcat:Datasets, dcat:DataServices and dcat:DatasetSeries
     */
    void convertDataResource(Element metadata, Resource dataset) {
        final String id = getTextOrNull(metadata.selectSingleNode("gmd:fileIdentifier/gco:CharacterString"));

        final String title = getTextOrNull(metadata.selectSingleNode("gmd:identificationInfo/*/gmd:citation/gmd:CI_Citation/gmd:title/gco:CharacterString"));
        if (title != null) {

            if (title.contains("#locale-eng:")) {
                // In some source systems, German and English titles are strangely combined in one text.
                String germanTitle = StringUtils.substringBefore(title, "#locale-eng:");
                String englishTitle = StringUtils.substringAfter(title, "#locale-eng:");
                dataset.addLiteral(DCTerms.title, model.createLiteral(germanTitle.trim(), LANGUAGE_DE));
                dataset.addLiteral(DCTerms.title, model.createLiteral(englishTitle.trim(), LANGUAGE_EN));
            } else {
                dataset.addLiteral(DCTerms.title, model.createLiteral(title.trim(), LANGUAGE_DE));
            }
        }

        final String description = getTextOrNull(metadata.selectSingleNode("gmd:identificationInfo/*/gmd:abstract/gco:CharacterString"));
        if (description != null) {
            dataset.addLiteral(DCTerms.description, model.createLiteral(description.trim(), LANGUAGE_DE));
        }

        final List<Node> descriptiveKeywordNodes = metadata.selectNodes("gmd:identificationInfo/*/gmd:descriptiveKeywords/gmd:MD_Keywords/gmd:keyword/gco:CharacterString");
        final Set<String> descriptiveKeywords = descriptiveKeywordNodes.stream().map(Node::getText).collect(Collectors.toSet());

        for (String keyword : descriptiveKeywords) {
            if (StringUtils.isNotEmpty(keyword)) {
                if (settings.keywordRemove == null || !settings.keywordRemove.contains(keyword)) {
                    keyword = keyword.replaceAll("\n", " "); // some people use linebreaks in keywords
                    dataset.addLiteral(DCAT.keyword, model.createLiteral(keyword, LANGUAGE_DE));
                }
            }
        }

        final Element languageCode = (Element) metadata.selectSingleNode("gmd:language/gmd:LanguageCode");
        if (languageCode != null) {
            dataset.addProperty(DCTerms.language, model.createResource(Mapping.mapLanguageCode(languageCode)));
        }

        final Element characterSet = (Element) metadata.selectSingleNode("gmd:characterSet/gmd:MD_CharacterSetCode");
        if (characterSet != null) {
            try {
                dataset.addLiteral(Content.characterEncoding,
                        model.createTypedLiteral(Mapping.mapCharacterSetCode(characterSet), XSDDatatype.XSDstring));
            } catch (IllegalArgumentException ignore) {
                log.warn("Dataset {} has an invalid characterEncoding.", id);
            }
        }

        final Element maintenanceFrequencyCode = (Element) metadata.selectSingleNode("gmd:identificationInfo/*/gmd:resourceMaintenance/gmd:MD_MaintenanceInformation/gmd:maintenanceAndUpdateFrequency/gmd:MD_MaintenanceFrequencyCode");
        if (maintenanceFrequencyCode != null) {
            try {
                dataset.addProperty(DCTerms.accrualPeriodicity, model.createResource(Mapping.mapMaintenanceFrequencyCode(maintenanceFrequencyCode)));
            } catch (IllegalArgumentException ignore) {
                log.warn("Dataset {} has an invalid maintenanceFrequencyCode.", id);
            }
        }

        final List<Node> topicCategories = metadata.selectNodes("gmd:identificationInfo/*/gmd:topicCategory/*");
        if (topicCategories.isEmpty()) {
            // The topic category is unset. Try to use the keywords to determine the category (DCAT theme) of the dataset.
            addThemesByKeywords(dataset, descriptiveKeywords);
        } else {
            addThemes(dataset, topicCategories);
        }


        if (metadata.selectSingleNode("gmd:identificationInfo/*/gmd:resourceConstraints/gmd:MD_LegalConstraints/*/gmx:Anchor[@xlink:href='http://inspire.ec.europa.eu/metadata-codelist/LimitationsOnPublicAccess/noLimitations']") != null) {
            dataset.addProperty(DCTerms.accessRights, model.createResource("http://inspire.ec.europa.eu/metadata-codelist/LimitationsOnPublicAccess/noLimitations"));
        }

        // creation, publication, modification etc.
        final Map<String, Element> dates = collectDates(metadata);
        final Element dateStamp = metadata.element("dateStamp");
        if (dateStamp != null) {
            dates.put("modified", dateStamp);
        }
        log.debug("Found {} dates.", dates.size());

        if (dates.containsKey("publication")) {
            addDate(dataset, DCTerms.issued, dates.get("publication"));
        } else if (dates.containsKey("creation")) {
            // 1st fallback for publication date: creation date
            addDate(dataset, DCTerms.issued, dates.get("creation"));
        } else if (dates.containsKey("modified")) {
            // 2nd fallback for publication date: modification date
            addDate(dataset, DCTerms.issued, dates.get("modified"));
        }
        if (dates.containsKey("modified")) {
            addDate(dataset, DCTerms.modified, dates.get("modified"));
        }


        // spatial extend
        final Element elementBoundingBox = (Element) metadata.selectSingleNode("gmd:identificationInfo/*/gmd:extent/gmd:EX_Extent/gmd:geographicElement/gmd:EX_GeographicBoundingBox");
        addBoundingBox(dataset, elementBoundingBox);

        // temporal extend
        final Element temporalExtend = (Element) metadata.selectSingleNode("gmd:identificationInfo/*/gmd:extent/gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent");
        addTemporalExtend(dataset, temporalExtend);

        if (settings.addMissingTemporalStart && !dataset.hasProperty(DCTerms.temporal)) {
            // The dataset has no temporal coverage. We repair it by adding the "issued" date as temporal start.
            final Resource periodOfTime = model.createResource();
            periodOfTime.addProperty(RDF.type, DCTerms.PeriodOfTime);
            periodOfTime.addLiteral(SchemaOrg.startDate, dataset.getProperty(DCTerms.issued).getLiteral());
            dataset.addProperty(DCTerms.temporal, periodOfTime);
        }

        // responsible parties
        final Map<String, Element> responsibleParties = collectResponsibleParties(metadata);

        if (responsibleParties.containsKey("publisher")) {
            dataset.addProperty(DCTerms.publisher, createFoafOrg(responsibleParties.get("publisher"), model));
        } else if (responsibleParties.containsKey("pointOfContact")) {
            // There is no publisher given. Therefore, we use the pointOfContact instead.
            dataset.addProperty(DCTerms.publisher, createFoafOrg(responsibleParties.get("pointOfContact"), model));
        }

        if (responsibleParties.containsKey("pointOfContact")) {
            dataset.addProperty(DCAT.contactPoint, createVcardOrganization(responsibleParties.get("pointOfContact"), model));
        }

        // license
        final List<Node> otherLegalConstraints = metadata.selectNodes("gmd:identificationInfo/*/gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:otherConstraints/gco:CharacterString");
        addLicenseInformation(dataset, otherLegalConstraints);

        // gmd:referenceSystemInfo
        final List<Node> rsIdentifiers = metadata.selectNodes("gmd:referenceSystemInfo/gmd:MD_ReferenceSystem/gmd:referenceSystemIdentifier/gmd:RS_Identifier");
        for (Node n : rsIdentifiers) {
            final Element rsIdentifier = (Element) n;
            final Node anchor = rsIdentifier.selectSingleNode("gmd:code/gmx:Anchor[@xlink:href='urn:ogc:def:crs:EPSG']/text()");
            final Node anchorRef = rsIdentifier.selectSingleNode("gmd:code/gmx:Anchor/@xlink:href");
            final Node textNode = rsIdentifier.selectSingleNode("gmd:code/gco:CharacterString/text()");
            if (anchor != null) {
                dataset.addProperty(DCTerms.conformsTo, ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/" + anchor.getText()));
            } else if (anchorRef != null && anchorRef.getText().startsWith("http://www.opengis.net/def/crs/EPSG/")) {
                dataset.addProperty(DCTerms.conformsTo, ResourceFactory.createResource(anchorRef.getText()));
            } else if (textNode != null) {
                if (textNode.getText().startsWith("http://www.opengis.net/def/crs/EPSG/")) {
                    dataset.addProperty(DCTerms.conformsTo, ResourceFactory.createResource(textNode.getText()));
                } else if (NumberUtils.toInt(textNode.getText()) > 0) {
                    // The RS_Identifier element contains a number. This is probably an EPSG code.
                    dataset.addProperty(DCTerms.conformsTo, ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/" + textNode.getText()));
                } else {
                    log.warn("Unknown gmd:referenceSystemIdentifier value {}", textNode.getText());
                }
            }
        }

        // convert gmd:DQ_DomainConsistency into dct:conformsTo elements
        final List<Node> domainConsistency = metadata.selectNodes("gmd:dataQualityInfo/gmd:DQ_DataQuality/gmd:report/gmd:DQ_DomainConsistency/gmd:measureIdentification/gmd:MD_Identifier/gmd:code/gco:CharacterString");
        for (Node node : domainConsistency) {
            String identifier = node.getText();
            if ("VERORDNUNG (EG) Nr. 1089/2010 DER KOMMISSION vom 23. November 2010 zur Durchführung der Richtlinie 2007/2/EG des Europäischen Parlaments und des Rates hinsichtlich der Interoperabilität von Geodatensätzen und -diensten".equals(identifier)) {
                dataset.addProperty(DCTerms.conformsTo, model.createResource("http://data.europa.eu/eli/reg/2010/1089/oj"));
            } else if ("VERORDNUNG (EG) Nr. 976/2009 DER KOMMISSION vom 19. Oktober 2009 zur Durchführung der Richtlinie 2007/2/EG des Europäischen Parlaments und des Rates hinsichtlich der Netzdienste".equals(identifier)) {
                dataset.addProperty(DCTerms.conformsTo, model.createResource("http://data.europa.eu/eli/reg/2009/976/oj"));
            } else if (StringUtils.contains(identifier, "Verordnung (EG) Nr. 1205/2008")) {
                dataset.addProperty(DCTerms.conformsTo, model.createResource("http://data.europa.eu/eli/reg/2008/1205/oj"));
            }
        }

        // convert INSPIRE keywords into dcatap:applicableLegislation properties
        if (descriptiveKeywords.contains("inspireidentifiziert")) {
            dataset.addProperty(DCATAP.applicableLegislation, model.createResource("http://data.europa.eu/eli/dir/2007/2/2019-06-26"));
        }

        // Create the HVD properties in accordance with the GDI-DE mapping rules of 2024-03-20
        final List<Node> hvdRelevantKeywordNodes =
                metadata.selectNodes("gmd:identificationInfo/*/gmd:descriptiveKeywords/gmd:MD_Keywords[" +
                        "gmd:thesaurusName/gmd:CI_Citation/gmd:title/gco:CharacterString = 'High-value dataset categories'" +
                        " or gmd:thesaurusName/gmd:CI_Citation/gmd:title/gmx:Anchor/@xlink:href = 'http://data.europa.eu/bna/asd487ae75']");
        if (!hvdRelevantKeywordNodes.isEmpty()) {
            dataset.addProperty(DCATAP.applicableLegislation, model.createResource("http://data.europa.eu/eli/reg_impl/2023/138/oj"));
        }
        for (Node keyword : hvdRelevantKeywordNodes) {
            Node characterString = keyword.selectSingleNode("./gmd:keyword/gco:CharacterString/text()");
            Node anchorRef = keyword.selectSingleNode("./gmd:keyword/gmx:Anchor/@xlink:href");

            if (characterString != null) {
                try {
                    dataset.addProperty(DCATAP.hvdCategory, Mapping.mapHvdCategory(characterString.getStringValue()));
                } catch (IllegalArgumentException e) {
                    log.warn("Invalid HVD category {} for dataset {}", characterString.getStringValue(), id);
                }
            }

            if (anchorRef != null) {
                try {
                    dataset.addProperty(DCATAP.hvdCategory, model.createResource(anchorRef.getStringValue()));
                } catch (IllegalArgumentException e) {
                    log.warn("Invalid HVD category {} for dataset {}", anchorRef.getStringValue(), id);
                }
            }
        }

        // convert preview image
        final Node graphicOverview = metadata.selectSingleNode("gmd:identificationInfo/*/gmd:graphicOverview/gmd:MD_BrowseGraphic/gmd:fileName/gco:CharacterString");
        if (graphicOverview != null) {
            String url = graphicOverview.getText().trim();
            if (isValidURLandIRI(url)) {
                dataset.addProperty(FOAF.thumbnail, model.createResource(url));
            }
        }

        // some fixed properties
        dataset.addProperty(DCATAPde.contributorID, model.createResource(settings.contributorId));

    }

    static boolean  isDatasetSeries(Element metadata) {
        final Element scopeCode = (Element) metadata.selectSingleNode("gmd:hierarchyLevel/gmd:MD_ScopeCode");
        if (scopeCode != null && "series".equals(scopeCode.attributeValue("codeListValue"))) {
            return true;
        }

        final String title = getTextOrNull(metadata.selectSingleNode("gmd:identificationInfo/*/gmd:citation/gmd:CI_Citation/gmd:title/gco:CharacterString"));
        return StringUtils.endsWith(title, "(Serie)");
    }

    static boolean isDataService(Element metadata) {
        final Element scopeCode = (Element) metadata.selectSingleNode("gmd:hierarchyLevel/gmd:MD_ScopeCode");
        return scopeCode != null && "service".equals(scopeCode.attributeValue("codeListValue"));
    }

    /**
     * Convert a gmd:MD_Metadata {@link Element} into a DCAT-AP.de compliant dcat:Dataset {@link Resource}.
     */
    public Resource convert(Element metadata) {

        if ("GetRecordByIdResponse".equals(metadata.getName())) {
            metadata = metadata.element("MD_Metadata");
        }

        if (!"MD_Metadata".equals(metadata.getName())) {
            throw new IllegalArgumentException("Only a gmd:MD_Metadata element can be converted.");
        }

        addNamespaces(metadata);

        final Element scopeCode = (Element) metadata.selectSingleNode("gmd:hierarchyLevel/gmd:MD_ScopeCode");
        if (scopeCode == null) {
            log.warn("MD_Metadata without MD_ScopeCode.");
        } else if ( isDataService(metadata) ){
            return null;
        }

        final Resource dataset;
        final String id = getTextOrNull(metadata.selectSingleNode("gmd:fileIdentifier/gco:CharacterString"));
        if (id != null) {
            dataset = model.createResource(settings.baseIRI + id);
            dataset.addLiteral(Adms.identifier, id.trim());
            dataset.addLiteral(DCTerms.identifier, id.trim());
        } else {
            dataset = model.createResource();
        }

        dataset.addProperty(RDF.type, DCAT.Dataset);

        if (scopeCode != null) {
            try {
                dataset.addProperty(DCTerms.type, model.createResource(Mapping.mapScopeCode(scopeCode)));
            } catch (IllegalArgumentException ignore) {
                log.warn("Dataset {} has an invalid scopeCode.", id);
            }
        }

        convertDataResource(metadata, dataset);

        final Element mdIdentifier = (Element) metadata.selectSingleNode("gmd:identificationInfo/*/gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:MD_Identifier/gmd:code/gco:CharacterString");
        final String searchId = mdIdentifier != null ? mdIdentifier.getTextTrim() : id;

        LocalDate datasetModificationDate = RDFUtils.literalToDate(dataset.getProperty(DCTerms.modified).getLiteral());

        // add coupled services
        if (cswInterface != null && searchId != null) {
            List<Element> services;
            try {
                services = cswInterface.findOperatesOn(searchId);
            } catch (Exception e) {
                log.error("Error searching for coupled services of dataset {}: {}", id, e.getMessage());
                services = Collections.emptyList();
            }
            for (Element service : services) {

                createDataService(service, model);

                try {
                    for (Resource dist : convertServiceToDistributions(service)) {
                        dataset.addProperty(DCAT.distribution, dist);

                        // Is the modification date of the service more recent?
                        final Literal distributionModified = dist.getProperty(DCTerms.modified).getLiteral();
                        final LocalDate distributionModificationDate = RDFUtils.literalToDate(distributionModified);
                        if (distributionModificationDate.isAfter(datasetModificationDate)) {
                            // update dct:modified value of the dataset
                            dataset.removeAll(DCTerms.modified);
                            dataset.addProperty(DCTerms.modified, distributionModified);
                            datasetModificationDate = distributionModificationDate;
                        }
                    }
                } catch (IllegalArgumentException ex) {
                    log.warn("Skipping distribution for dataset {}: {}", searchId, ex.getMessage());
                }
            }
        }

        // distributions directly included in the dataset
        if (isMultiFormatDataset(metadata)) {
            // This is a multi-format dataset.
            log.debug("Multi-format dataset.");
            for (Resource distribution : convertMultiFormatDataset(metadata, false)) {
                dataset.addProperty(DCAT.distribution, distribution);
            }

        } else {
            final List<Node> onlineResourcesDownload = metadata.selectNodes("gmd:distributionInfo/*/gmd:transferOptions/*/gmd:onLine/*[gmd:function/*/@codeListValue='download' and gmd:linkage/*[text()]]");
            for (Node onlineResource : onlineResourcesDownload) {
                addDistribution(dataset, (Element) onlineResource);
            }
        }

        final List<Node> otherLegalConstraints = metadata.selectNodes("gmd:identificationInfo/*/gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:otherConstraints/gco:CharacterString");
        final String dateStamp = getTextOrNull(metadata.element("dateStamp"));

        final List<Node> onlineResourceInformation = metadata.selectNodes("gmd:distributionInfo/*/gmd:transferOptions/*/gmd:onLine/*[gmd:function/*/@codeListValue='information' and gmd:linkage/*[text()]]");
        for (Node node : onlineResourceInformation) {
            final Element linkage = ((Element) node).element("linkage");
            if (linkage != null) {
                final String url = repairURL(linkage.getStringValue().trim());
                if (settings.findWMSinInfo
                        && StringUtils.containsIgnoreCase(url, "service=WMS")
                        && StringUtils.containsIgnoreCase(url, "request=GetCapabilities")) {
                    Resource distribution = convertOnlineResourceToDistribution(node, otherLegalConstraints, dateStamp, null);
                    if (distribution != null)
                        dataset.addProperty(DCAT.distribution, distribution);
                } else {
                    if (isValidURLandIRI(url)) {
                        dataset.addProperty(model.createProperty(FOAF.NS, "page"), model.createResource(url));
                    } else {
                        log.info("Invalid info page {} for dataset {}", url, id);
                    }
                }
            }
        }

        if (settings.ignoreSeriesWithoutDistribution && isDatasetSeries(metadata) && !dataset.hasProperty(DCAT.distribution)) {
            // Geo series might not have a download option or a couple service (which is totally valid). If configured,
            // this kind of series will be ignored.
            dataset.removeProperties();
            return null;
        }

        return dataset;
    }

    /**
     * Create a dcat:DataService resource for a service.
     */
    Resource createDataService(Element metadata, Model model) {
        final Resource dataService;
        addNamespaces(metadata);
        final String id = getTextOrNull(metadata.selectSingleNode("gmd:fileIdentifier/gco:CharacterString"));
        if (id != null) {
            dataService = model.createResource(settings.baseIRI + id);
            dataService.addLiteral(Adms.identifier, id.trim());
            dataService.addLiteral(DCTerms.identifier, id.trim());
        } else {
            dataService = model.createResource();
        }

        dataService.addProperty(RDF.type, DCAT.DataService);

        convertDataResource(metadata, dataService);

        String url = getServiceEndpoint(metadata);

        if (StringUtils.contains(StringUtils.lowerCase(url), "request=getcapabilities")) {
            // the GetCapabilites request is linked
            dataService.addProperty(DCAT.endpointDescription, model.createResource(url));
            dataService.addProperty(DCAT.endpointURL, model.createResource(StringUtils.substringBefore(url, "?")));
        } else {
            dataService.addProperty(DCAT.endpointURL, model.createResource(url));
        }

        List<Node> linkedDatasets = metadata.selectNodes("gmd:identificationInfo[1]/*/srv:operatesOn/@xlink:href");
        for (Node linkedDataset : linkedDatasets) {
            dataService.addProperty(DCAT.servesDataset, model.createResource(StringUtils.trim(linkedDataset.getText())));
        }

        return dataService;
    }

    private boolean isMultiFormatDataset(Element metadata) {
        final List<Node> formatElements = metadata.selectNodes("gmd:distributionInfo/*/gmd:distributionFormat");
        return formatElements.size() > 1;
    }

    private void addDate(Resource dataset, Property property, Element timeElement) {
        String value = getDateValue(timeElement);
        if (timeElement.element("Date") != null) {
            dataset.addLiteral(property, model.createTypedLiteral(value, XSDDatatype.XSDdate));
        } else if (timeElement.element("DateTime") != null) {
            dataset.addLiteral(property, model.createTypedLiteral(value, XSDDatatype.XSDdateTime));
        } else {
            dataset.addLiteral(property, model.createLiteral(value));
        }
    }

    private String getDateValue(Element timeElement) {
        final String value;
        if (timeElement.element("Date") != null) {
            value = timeElement.elementText("Date");
        } else if (timeElement.element("DateTime") != null) {
            value = timeElement.elementText("DateTime");
        } else if (timeElement.elements().size() > 0) {
            value = timeElement.elements().get(0).getText();
        } else {
            value = StringUtils.trimToEmpty(timeElement.getStringValue());
        }
        return value;
    }

    private static String getGlobalFormatForAllTransferOptions(Element onlineResource) {
        final Element distributionElement = onlineResource.getParent().getParent().getParent().getParent();
        if (!"MD_Distribution".equals(distributionElement.getName())) {
            throw new IllegalArgumentException("Could not navigate to MD_Distribution element.");
        }
        final Element distributionFormat = distributionElement.element("distributionFormat");

        if( distributionFormat == null)
            return null;

        return getTextOrNull(distributionFormat.selectSingleNode("gmd:MD_Format/gmd:name/gco:CharacterString"));
    }

    void addDistribution(Resource dataset, Element onlineResource) {
        final Model model = dataset.getModel();


        // there might be a single format specified or all transfer options
        final String globalFormat = getGlobalFormatForAllTransferOptions(onlineResource);

        final String url = repairURL(getTextOrNull(onlineResource.selectSingleNode("gmd:linkage/gmd:URL")));
        final String name = getTextOrNull(onlineResource.selectSingleNode("gmd:name/*"));
        final String description = getTextOrNull(onlineResource.selectSingleNode("gmd:description/*"));
        final String applicationProfile = getTextOrNull(onlineResource.selectSingleNode("gmd:applicationProfile/*"));

        // Check if the specified URL is a valid URI and IRI.
        if (!isValidURLandIRI(url)) {
            log.info("Distribution has invalid URL {}", url);
            return;
        }

        final Resource distribution = model.createResource(url);
        distribution.addProperty(RDF.type, DCAT.Distribution);
        distribution.addProperty(DCAT.downloadURL, model.createResource(url));
        distribution.addProperty(DCAT.accessURL, model.createResource(url));

        Resource formatResource;

        // Now it seems to be best to evaluate the transfer options' gmd:applicationProfile first.
        if (StringUtils.isNotEmpty(applicationProfile)) {
            formatResource = Mapping.mapApplicationProfile(applicationProfile);
        } else {
            try {
                formatResource = Mapping.mapFormat(globalFormat);
            } catch (IllegalArgumentException e) {
                formatResource = null;
            }
            if (formatResource == null) {
                formatResource = guessFormat(url);
            }
        }

        if (formatResource != null) {
            distribution.addProperty(DCTerms.format, formatResource);
        } else if (!ignoreInvalidMapping) {
            throw new RuntimeException("Could not determine format of distribution " + url);
        }

        if (StringUtils.isNotEmpty(name)) {
            distribution.addLiteral(DCTerms.title, name);
        }
        if (StringUtils.isNotEmpty(description)) {
            distribution.addLiteral(DCTerms.description, description);
        }

        // Copy license information from dataset if available
        if (dataset.hasProperty(DCTerms.license)) {
            distribution.addProperty(DCTerms.license, dataset.getProperty(DCTerms.license).getResource());
        }
        if (dataset.hasProperty(DCATAPde.licenseAttributionByText)) {
            distribution.addLiteral(DCATAPde.licenseAttributionByText, dataset.getProperty(DCATAPde.licenseAttributionByText).getLiteral());
        }

        dataset.addProperty(DCAT.distribution, distribution);
    }

    /**
     * Iterate through the content of gmd:otherConstraints to find license information in JSON
     *
     * @return JSONObject with license information or <code>null</code> if no license information in JSON can be found
     */
    JSONObject findLicenseInformation(List<Node> otherLegalConstraints) {
        for (Node otherLegalConstraint : otherLegalConstraints) {
            try {
                return new JSONObject(new JSONTokener(otherLegalConstraint.getText()));
            } catch (JSONException ignore) {
                // this text is not JSON and therefore does not contain the requested information.
            }
        }
        return null;
    }

    /**
     * Add dcat:theme properties to the dataset, based on the specified  topicCategories.
     */
    void addThemes(Resource dataset, List<Node> topicCategories) {
        final Model model = dataset.getModel();
        for (Node node : topicCategories) {
            final String category = node.getText();
            for (final String theme : themeMapping.keySet()) {
                if (ArrayUtils.contains(themeMapping.get(theme), category)) {
                    dataset.addProperty(DCAT.theme, model.createResource("http://publications.europa.eu/resource/authority/data-theme/" + theme));
                }
            }
        }
    }


    /**
     * Use the dataset's keywords to determine the category (DCAT theme) of the dataset.
     *
     * @param dataset             the {@link Resource} of the resulting DCAT dataset
     * @param descriptiveKeywords Set of keywords
     */
    private void addThemesByKeywords(Resource dataset, Set<String> descriptiveKeywords) {
        for (String keyword : descriptiveKeywords) {
            for (String theme : themeMapping.keySet()) {
                if (ArrayUtils.contains(themeMapping.get(theme), keyword)) {
                    dataset.addProperty(DCAT.theme, model.createResource("http://publications.europa.eu/resource/authority/data-theme/" + theme));
                }
            }

            if (keywordMapping.containsKey(keyword)) {
                final String theme = keywordMapping.get(keyword);
                dataset.addProperty(DCAT.theme, model.createResource("http://publications.europa.eu/resource/authority/data-theme/" + theme));
            }
        }
    }

}
