package de.landsh.opendata.csw2dcat;

import de.landsh.opendata.csw2dcat.vocabulary.*;
import org.apache.jena.datatypes.xsd.XSDDatatype;
import org.apache.jena.rdf.model.*;
import org.apache.jena.vocabulary.DCAT;
import org.apache.jena.vocabulary.DCTerms;
import org.apache.jena.vocabulary.RDF;
import org.apache.jena.vocabulary.VCARD4;
import org.dom4j.*;
import org.dom4j.io.SAXReader;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.io.IOException;
import java.io.StringReader;
import java.io.Writer;
import java.util.*;

import static org.junit.jupiter.api.Assertions.*;

public class MDMetadata2DatasetTests {

    private static final String BASE_IRI = "http://example.org/dataset/";
    private static final String CONTRIBUTOR_ID = "http://example.org/myContributorId";
    private static final Resource FILE_TYPE_WMS_SRC = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/WMS_SRVC");
    private static final Resource FILE_TYPE_ATOM = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/ATOM");
    private static final Resource FILE_TYPE_HTML = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/HTML");
    private static final Resource FILE_TYPE_WFS_SRC = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/WFS_SRVC");
    private static final Resource FILE_TYPE_CSV = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/CSV");
    private static final Resource FILE_TYPE_SHP = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/SHP");
    private static final Resource FILE_TYPE_GML = ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/GML");

    private static final Property PROPERTY_applicable_Legislation = ResourceFactory.createProperty("http://data.europa.eu/r5r/", "applicableLegislation");
    static Namespace NS_GMD = Namespace.get("gmd", "http://www.isotc211.org/2005/gmd");
    final Model model = ModelFactory.createDefaultModel();
    private final SAXReader saxReader = new SAXReader();
    private MDMetadata2Dataset service;
    private CswInterface cswInterface;
    private BridgeSettings settings;

    private static void checkSeehundeDistributions(Map<String, Resource> distributionMap) {
        assertEquals(6, distributionMap.size());

        Resource csv = distributionMap.get("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=WFS&version=2.0.0&request=GetFeature&typeName=Seehunde_Liegeplaetze_ab1989&outputFormat=CSV");
        Resource shape = distributionMap.get("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=WFS&version=2.0.0&request=GetFeature&typeName=Seehunde_Liegeplaetze_ab1989&outputFormat=SHAPE-ZIP");
        Resource wms = distributionMap.get("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=wms&version=1.3.0&request=GetCapabilities");
        Resource wfs = distributionMap.get("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=wfs&version=2.0.0&request=GetCapabilities");
        Resource shapeExample = distributionMap.get("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=WFS&version=2.0.0&request=GetFeature&typeName=Seehunde_Liegeplaetze_ab1989&outputFormat=SHAPE-ZIP&CQL_FILTER=ROBBENJAHR%3D2016");
        Resource csvExample = distributionMap.get("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=WFS&version=2.0.0&request=GetFeature&typeName=Seehunde_Liegeplaetze_ab1989&outputFormat=CSV&CQL_FILTER=ROBBENJAHR%3D2016");

        assertNotNull(csv);
        assertNotNull(shape);
        assertNotNull(wms);
        assertNotNull(wfs);
        assertNotNull(shapeExample);
        assertNotNull(csvExample);

        assertEquals(FILE_TYPE_SHP, shape.getProperty(DCTerms.format).getResource());
        assertEquals(FILE_TYPE_SHP, shapeExample.getProperty(DCTerms.format).getResource());
        assertEquals(FILE_TYPE_CSV, csv.getProperty(DCTerms.format).getResource());
        assertEquals(FILE_TYPE_CSV, csvExample.getProperty(DCTerms.format).getResource());
        assertEquals(FILE_TYPE_WFS_SRC, wfs.getProperty(DCTerms.format).getResource());
        assertEquals(FILE_TYPE_WMS_SRC, wms.getProperty(DCTerms.format).getResource());

        assertEquals(ResourceFactory.createResource("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=wms&version=1.3.0&request=GetCapabilities"), wms.getProperty(DCAT.accessURL).getResource());
        assertEquals(ResourceFactory.createResource("http://mdi-sh.org/geoserver_lkn/Seehunde_Liegeplaetze/ows?service=wms&version=1.3.0&request=GetCapabilities"), wms.getProperty(DCAT.downloadURL).getResource());
        assertTrue(wms.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(wms.hasLiteral(DCATAPde.licenseAttributionByText, "LKN SH Nationalparkverwaltung"));
        assertEquals("Download WFS: csv-file", csv.getProperty(DCTerms.title).getString());
    }

    @BeforeEach
    public void setUp() {
        cswInterface = Mockito.mock(CswInterface.class);

        settings = new BridgeSettings();
        settings.contributorId = CONTRIBUTOR_ID;
        settings.baseIRI = BASE_IRI;

        service = new MDMetadata2Dataset(ModelFactory.createDefaultModel(), cswInterface, settings);

        service.ignoreInvalidMapping = false;
    }

    @Test
    public void testConvert() throws DocumentException, IOException {
        model.read(getClass().getResourceAsStream("/f7f90143-c2ad-46b2-934d-93dfd5e0f031.rdf"), null);
        final Resource referenceDataset = model.listResourcesWithProperty(RDF.type, DCAT.Dataset).next();

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/f7f90143-c2ad-46b2-934d-93dfd5e0f031.xml"));

        final Resource result = service.convert(inputDocument);

        assertEquals("Straßenschlüsselverzeichnis der Stadt Elmshorn ohne Koordinatenbezug", result.getProperty(DCTerms.title).getString());
        assertEquals("Straßenschlüsselverzeichnis der Stadt Elmshorn mit Gemeindeschlüssel und Postleitzahl ohne Koordinatenbezug", result.getProperty(DCTerms.description).getString());
        assertEquals("f7f90143-c2ad-46b2-934d-93dfd5e0f031", result.getProperty(DCTerms.identifier).getString());
        assertEquals("http://publications.europa.eu/resource/authority/language/DEU", result.getProperty(DCTerms.language).getResource().getURI());
        assertEquals("f7f90143-c2ad-46b2-934d-93dfd5e0f031", result.getProperty(Adms.identifier).getString());
        assertEquals(CONTRIBUTOR_ID, result.getProperty(DCATAPde.contributorID).getResource().getURI());
        assertEquals("utf8", result.getProperty(Content.characterEncoding).getString());
        assertEquals("http://inspire.ec.europa.eu/metadata-codelist/ResourceType/series", result.getProperty(DCTerms.type).getResource().getURI());
        assertEquals("http://publications.europa.eu/resource/authority/frequency/ANNUAL", result.getProperty(DCTerms.accrualPeriodicity).getResource().getURI());
        assertEquals("http://inspire.ec.europa.eu/metadata-codelist/LimitationsOnPublicAccess/noLimitations", result.getProperty(DCTerms.accessRights).getResource().getURI());
        RdfTester.assertStatementPresent(result, DCAT.theme, referenceDataset);
        RdfTester.assertStatementPresent(result, DCAT.keyword, referenceDataset);

        assertTrue(result.hasProperty(DCTerms.modified));
        assertEquals("http://www.w3.org/2001/XMLSchema#date", result.getProperty(DCTerms.modified).getLiteral().getDatatypeURI());
        assertEquals("2022-05-16", result.getProperty(DCTerms.modified).getLiteral().getString());

        assertTrue(result.hasProperty(DCTerms.issued));
        assertEquals("http://www.w3.org/2001/XMLSchema#date", result.getProperty(DCTerms.issued).getLiteral().getDatatypeURI());
        assertEquals("2022-05-16", result.getProperty(DCTerms.issued).getLiteral().getString());

        // spacial extend
        assertTrue(result.hasProperty(DCTerms.spatial));
        final Resource location = result.getProperty(DCTerms.spatial).getResource();
        assertTrue(location.hasProperty(RDF.type, DCTerms.Location));
        assertTrue(location.hasProperty(Locn.geometry));
        assertEquals("http://www.opengis.net/ont/geosparql#wktLiteral", location.getProperty(Locn.geometry).getLiteral().getDatatypeURI());
        assertEquals("POLYGON((5.81 54.76,15.77 54.76,15.77 47.26,5.81 47.26,5.81 54.76))", location.getProperty(Locn.geometry).getString());

        // publisher
        assertTrue(result.hasProperty(DCTerms.publisher));
        final Resource publisher = result.getProperty(DCTerms.publisher).getResource();
        assertTrue(publisher.hasProperty(RDF.type, FOAF.Organization));
        assertEquals("Stadt Elmshorn", publisher.getProperty(FOAF.name).getString());
        assertEquals("mailto:gis@elmshorn.de", publisher.getProperty(FOAF.mbox).getResource().getURI());
        assertEquals("tel:+49-4121-231-446", publisher.getProperty(FOAF.phone).getResource().getURI());
        assertEquals("http://www.elmshorn.de", publisher.getProperty(FOAF.homepage).getResource().getURI());
        assertTrue(publisher.hasProperty(Locn.address));
        {
            final Resource address = publisher.getProperty(Locn.address).getResource();
            assertTrue(address.hasProperty(RDF.type, Locn.Address));
            assertEquals("Schulstraße 15 - 17", address.getProperty(Locn.thoroughfare).getLiteral().getString());
            assertEquals("Elmshorn", address.getProperty(Locn.postName).getLiteral().getString());
            assertEquals("25335", address.getProperty(Locn.postCode).getLiteral().getString());
            assertEquals("Schleswig-Holstein", address.getProperty(Locn.adminUnitL2).getLiteral().getString());
            assertEquals("Deutschland", address.getProperty(Locn.adminUnitL1).getLiteral().getString());
        }

        assertTrue(result.hasProperty(DCAT.contactPoint));
        final Resource contactPoint = result.getProperty(DCAT.contactPoint).getResource();
        {
            final Resource address = contactPoint.getProperty(VCARD4.hasAddress).getResource();
            assertEquals("Schulstraße 15 - 17", address.getProperty(VCARD4.street_address).getLiteral().getString());
            assertEquals("Elmshorn", address.getProperty(VCARD4.locality).getLiteral().getString());
            assertEquals("25335", address.getProperty(VCARD4.postal_code).getLiteral().getString());
            assertEquals("Schleswig-Holstein", address.getProperty(VCARD4.region).getLiteral().getString());
            assertEquals("Deutschland", address.getProperty(VCARD4.country_name).getLiteral().getString());
        }
        final Resource telephone = contactPoint.getProperty(VCARD4.hasTelephone).getResource();
        assertTrue(telephone.hasProperty(RDF.type, VCARD4.Voice));
        assertEquals("tel:+49-4121-231-446", telephone.getProperty(VCARD4.hasValue).getResource().getURI());
    }

    @Test
    public void testConvert_coupledServices() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/7164d86e-8c55-4e08-ab1f-601b3002d1d1.xml"));

        final Resource result = service.convert(inputDocument);
        final Model model = result.getModel();

        assertTrue(result.hasLiteral(DCTerms.identifier, "7164d86e-8c55-4e08-ab1f-601b3002d1d1"));
        assertTrue(result.hasProperty(DCAT.theme, model.createResource("http://publications.europa.eu/resource/authority/data-theme/GOVE")));
        assertTrue(result.hasProperty(DCAT.theme, model.createResource("http://publications.europa.eu/resource/authority/data-theme/REGI")));

        assertTrue(result.hasProperty(DCTerms.issued));
        assertEquals("\"2021-03-19\"^^xsd:date", result.getProperty(DCTerms.issued).getLiteral().toString());
        assertTrue(result.hasProperty(DCTerms.modified));
        assertEquals("\"2022-05-31\"^^xsd:date", result.getProperty(DCTerms.modified).getLiteral().toString());
        assertTrue(result.hasProperty(DCTerms.license, model.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(result.hasLiteral(DCATAPde.licenseAttributionByText, "© GDI-SH"));
    }

    private void setNsPrefixes(Model model) {
        model.setNsPrefix("dcat", DCAT.NS);
        model.setNsPrefix("dct", DCTerms.NS);
        model.setNsPrefix("vcard", VCARD4.NS);
        model.setNsPrefix("dcatde", "http://dcat-ap.de/def/dcatde/");
        model.setNsPrefix("adms", "http://www.w3.org/ns/adms#");
        model.setNsPrefix("cnt", "http://www.w3.org/2011/content#");
        model.setNsPrefix("geosparql", "http://www.opengis.net/ont/geosparql#");
        model.setNsPrefix("locn", "http://www.w3.org/ns/locn#");
        model.setNsPrefix("foaf", "http://xmlns.com/foaf/0.1/");
    }

    @Test
    public void convertServiceToDistributions() throws DocumentException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/service1.xml"));
        final Collection<Resource> result = service.convertServiceToDistributions(inputDocument.getRootElement());

        assertNotNull(result);
        assertEquals(1, result.size());

        final Resource distribution = result.iterator().next();

        assertTrue(distribution.hasProperty(RDF.type, DCAT.Distribution));
        assertTrue(distribution.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(distribution.hasProperty(DCTerms.format, FILE_TYPE_WMS_SRC));
        assertTrue(distribution.hasLiteral(DCATAPde.licenseAttributionByText, "© GDI-SH"));
        assertEquals("2022-05-31", distribution.getProperty(DCTerms.modified).getLiteral().getString());
        assertEquals("https://service.gdi-sh.de/WMS_SH_FD_Chronologen_KF?Service=wms&version=1.3.0&request=GetCapabilities",
                distribution.getProperty(DCAT.accessURL).getResource().getURI());
    }

    @Test
    public void testConvert_seegras() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/seegras.xml"));

        final Resource result = service.convert(inputDocument);
        assertTrue(result.hasProperty(DCTerms.spatial));
        assertTrue(result.hasProperty(DCTerms.temporal));
        Resource temporal = result.getProperty(DCTerms.temporal).getObject().asResource();
        assertTrue(temporal.hasProperty(SchemaOrg.startDate));
        assertTrue(temporal.hasProperty(SchemaOrg.endDate));
    }

    @Test
    public void testConvert_OnlineResource() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/6afcc783-3698-4412-8165-b67763d38930.xml"));

        final Resource result = service.convert(inputDocument);

//        final Model model = result.getModel();
//        setNsPrefixes(model);
//        model.write(System.out, "RDF/XML-ABBREV");
//        System.out.flush();

        assertEquals("http://dcat-ap.de/def/licenses/dl-by-de/2.0", result.getProperty(DCTerms.license).getResource().getURI());
        assertEquals("LKN SH Nationalparkverwaltung", result.getProperty(DCATAPde.licenseAttributionByText).getLiteral().getString());

        // Check publisher
        assertTrue(result.hasProperty(DCTerms.publisher));
        final Resource publisher = result.getProperty(DCTerms.publisher).getResource();
        assertTrue(publisher.hasProperty(RDF.type, FOAF.Organization));
        assertEquals("LKN, Nationalparkverwaltung schleswig-holsteinisches Wattenmeer", publisher.getProperty(FOAF.name).getLiteral().getString());
        assertEquals("mailto:nationalpark@lkn.landsh.de", publisher.getProperty(FOAF.mbox).getResource().getURI());
        assertEquals("tel:+49-48616160", publisher.getProperty(FOAF.phone).getResource().getURI());
        assertTrue(publisher.hasProperty(Locn.address));
        final Resource address1 = publisher.getProperty(Locn.address).getResource();
        assertTrue(address1.hasProperty(RDF.type, Locn.Address));
        assertEquals("Schlossgarten 1", address1.getProperty(Locn.thoroughfare).getLiteral().getString());
        assertEquals("Tönning", address1.getProperty(Locn.postName).getLiteral().getString());
        assertEquals("25832", address1.getProperty(Locn.postCode).getLiteral().getString());

        // Check contact point
        Resource contactPoint = result.getProperty(DCAT.contactPoint).getResource();
        Resource address = contactPoint.getProperty(VCARD4.hasAddress).getResource();
        assertEquals("Schlossgarten 1", address.getProperty(VCARD4.street_address).getLiteral().getString());
        assertEquals("Tönning", address.getProperty(VCARD4.locality).getLiteral().getString());
        assertEquals("25832", address.getProperty(VCARD4.postal_code).getLiteral().getString());

        Resource telephone = contactPoint.getProperty(VCARD4.hasTelephone).getResource();
        assertTrue(telephone.hasProperty(RDF.type, VCARD4.Voice));
        assertEquals("tel:+49-4861-616-46", telephone.getProperty(VCARD4.hasValue).getResource().getURI());

        assertTrue(result.hasProperty(DCAT.distribution));
        Resource distribution = result.getProperty(DCAT.distribution).getResource();
        assertEquals("https://s-h.nokis.org/Geodaten/gk/mod100/cv_jura/extra/ffh-evg-NATURA2000/NATURA2000SHArcExport2001/egv", distribution.getURI());
        assertEquals(FILE_TYPE_SHP, distribution.getProperty(DCTerms.format).getResource());
        assertEquals("http://dcat-ap.de/def/licenses/dl-by-de/2.0", distribution.getProperty(DCTerms.license).getResource().getURI());
        assertEquals("LKN SH Nationalparkverwaltung", distribution.getProperty(DCATAPde.licenseAttributionByText).getLiteral().getString());
        assertEquals("https://s-h.nokis.org/Geodaten/gk/mod100/cv_jura/extra/ffh-evg-NATURA2000/NATURA2000SHArcExport2001/egv", distribution.getProperty(DCAT.accessURL).getResource().getURI());
        assertEquals("https://s-h.nokis.org/Geodaten/gk/mod100/cv_jura/extra/ffh-evg-NATURA2000/NATURA2000SHArcExport2001/egv", distribution.getProperty(DCAT.downloadURL).getResource().getURI());

        assertEquals("http://www.w3.org/2001/XMLSchema#date", result.getProperty(DCTerms.modified).getLiteral().getDatatypeURI());
        assertEquals("2022-05-05", result.getProperty(DCTerms.modified).getLiteral().getString());

        assertEquals("http://www.w3.org/2001/XMLSchema#dateTime", result.getProperty(DCTerms.issued).getLiteral().getDatatypeURI());
        assertEquals("2002-02-12T00:00:00.000+01:00", result.getProperty(DCTerms.issued).getLiteral().getString());
    }

    @Test
    public void testConvert_triberg() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/triberg.xml"));

        service.ignoreInvalidMapping = true;

        final Resource result = service.convert(inputDocument);

//        final Model model = result.getModel();
//        setNsPrefixes(model);
//        model.write(System.out, "RDF/XML-ABBREV");
//        System.out.flush();

        assertEquals("62c99f78-b4fd-4773-b602-da56417668fd", result.getProperty(DCTerms.identifier).getString());

        // Check publisher
        assertTrue(result.hasProperty(DCTerms.publisher));
        final Resource publisher = result.getProperty(DCTerms.publisher).getResource();
        assertTrue(publisher.hasProperty(RDF.type, FOAF.Organization));
        assertEquals("Stadt Triberg", publisher.getProperty(FOAF.name).getLiteral().getString());
        assertEquals("mailto:Stadtverwaltung@triberg.de", publisher.getProperty(FOAF.mbox).getResource().getURI());
        assertEquals("tel:+49-7722-953-0", publisher.getProperty(FOAF.phone).getResource().getURI());
        assertTrue(publisher.hasProperty(Locn.address));
        final Resource address1 = publisher.getProperty(Locn.address).getResource();
        assertTrue(address1.hasProperty(RDF.type, Locn.Address));
        assertEquals("Hauptstraße 57", address1.getProperty(Locn.thoroughfare).getLiteral().getString());
        assertEquals("Triberg im Schwarzwald", address1.getProperty(Locn.postName).getLiteral().getString());
        assertEquals("78098", address1.getProperty(Locn.postCode).getLiteral().getString());

        // Check contact point
        final Resource contactPoint = result.getProperty(DCAT.contactPoint).getResource();
        final Resource address = contactPoint.getProperty(VCARD4.hasAddress).getResource();
        assertEquals("Hauptstraße 57", address.getProperty(VCARD4.street_address).getLiteral().getString());
        assertEquals("Triberg im Schwarzwald", address.getProperty(VCARD4.locality).getLiteral().getString());
        assertEquals("78098", address.getProperty(VCARD4.postal_code).getLiteral().getString());

        final Resource telephone = contactPoint.getProperty(VCARD4.hasTelephone).getResource();
        assertTrue(telephone.hasProperty(RDF.type, VCARD4.Voice));
        assertEquals("tel:+49-7722-953-0", telephone.getProperty(VCARD4.hasValue).getResource().getURI());

        assertTrue(result.hasProperty(DCAT.distribution));
        final Resource distribution = result.getProperty(DCAT.distribution).getResource();
        assertEquals("https://geodienste.komm.one/ows/services/org.250.b7ac18ab-368c-4a66-baf6-370fecb90983_wfs?SERVICE=WFS&Request=GetCapabilities", distribution.getURI());
        assertEquals("https://geodienste.komm.one/ows/services/org.250.b7ac18ab-368c-4a66-baf6-370fecb90983_wfs?SERVICE=WFS&Request=GetCapabilities", distribution.getProperty(DCAT.accessURL).getResource().getURI());
        assertEquals("https://geodienste.komm.one/ows/services/org.250.b7ac18ab-368c-4a66-baf6-370fecb90983_wfs?SERVICE=WFS&Request=GetCapabilities", distribution.getProperty(DCAT.downloadURL).getResource().getURI());

        assertEquals("http://www.w3.org/2001/XMLSchema#date", result.getProperty(DCTerms.modified).getLiteral().getDatatypeURI());
        assertEquals("2021-07-23", result.getProperty(DCTerms.modified).getLiteral().getString());

        assertEquals("http://www.w3.org/2001/XMLSchema#date", result.getProperty(DCTerms.issued).getLiteral().getDatatypeURI());
        assertEquals("2021-01-08", result.getProperty(DCTerms.issued).getLiteral().getString());
    }

    /**
     * Even if the input document contains invalid URLs the conversion must not fail.
     */
    @Test
    public void testConvert_invalidDistributionURL() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/invalid_distribution_url.xml"));

        service.ignoreInvalidMapping = true;

        final Resource result = service.convert(inputDocument);

        final Model model = result.getModel();
        model.write(Writer.nullWriter(), "RDF/XML-ABBREV");

    }

    /**
     * This dataset contains strange empty dates for revision and publication.
     */
    @Test
    public void testConvert_emptyDates() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/52f86bc6-732d-448d-8f15-cd9b0ab5ad54.xml"));
        final Resource result = service.convert(inputDocument);

        assertEquals("http://www.w3.org/2001/XMLSchema#date", result.getProperty(DCTerms.modified).getLiteral().getDatatypeURI());
        assertEquals("2022-05-20", result.getProperty(DCTerms.modified).getLiteral().getString());

        assertEquals("http://www.w3.org/2001/XMLSchema#date", result.getProperty(DCTerms.issued).getLiteral().getDatatypeURI());
        assertEquals("2020-08-01", result.getProperty(DCTerms.issued).getLiteral().getString());
    }

    @Test
    public void testConvert_thumbnail() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/52f86bc6-732d-448d-8f15-cd9b0ab5ad54.xml"));
        final Resource result = service.convert(inputDocument);

        RDFNode node = result.getProperty(FOAF.thumbnail).getObject();

        assertTrue(node.isURIResource());
        assertEquals("https://sh-mis.gdi-sh.de/metadaten/Vorschaubilder/dtk50.png", node.asResource().getURI());
    }

    @Test
    public void testConvert_thumbnail2() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/e95c2bed-84c0-4e7c-a194-1568887fc355.xml"));
        final Resource result = service.convert(inputDocument);

        RDFNode node = result.getProperty(FOAF.thumbnail).getObject();

        assertTrue(node.isURIResource());
        assertEquals("http://s-h.nokis.org/nokis/files/records/e95c2bed-84c0-4e7c-a194-1568887fc355/Seehunde_Monitoring_Vorschau.png", node.asResource().getURI());
    }

    /**
     * This ISO metadata contains a blank gmd:graphicOverview. Therefore, no thumbnail property will be set.
     *
     * @throws Exception
     */
    @Test
    public void testConvert_thumbnail_empty() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/tile.xml"));
        final Resource result = service.convert(inputDocument);

        assertFalse(result.hasProperty(FOAF.thumbnail));
    }

    /**
     * This dataset has a <code>nil</code> format specification.
     */
    @Test
    public void convert_nilFormat() throws Exception {
        service.ignoreInvalidMapping = true;
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/74775d8f-23e1-4ac5-bd73-4c3a9ce366f4.xml"));
        service.convert(inputDocument);
    }

    /**
     * <pre>e95c2bed-84c0-4e7c-a194-1568887fc355</pre> is a dataset.
     * <pre>5a9a19d0-f24b-427e-9bdd-c1c7da2e84e7</pre> is a coupled service of this dataset.
     */
    @Test
    public void convert_coupledServices_Seehunde() throws DocumentException, IOException {

        final Element serviceElement = saxReader.read(getClass().getResourceAsStream("/5a9a19d0-f24b-427e-9bdd-c1c7da2e84e7.xml")).getRootElement();

        Mockito.when(cswInterface.findOperatesOn("http://www.mdi-sh.org/#c365fcb1-f2b8-4e05-baf1-627b413da204")).thenReturn(Collections.singletonList(serviceElement));

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/e95c2bed-84c0-4e7c-a194-1568887fc355.xml"));

        final Resource result = service.convert(inputDocument);

        assertTrue(result.hasLiteral(DCTerms.identifier, "e95c2bed-84c0-4e7c-a194-1568887fc355"));

        // The modification date of the dataset itself is 2022-05-05. The modification date of the service is 2023-06-01.
        // The resulting modification date of the dataset is the latest modification date of one of its coupled services.
        assertEquals("\"2023-06-01\"^^xsd:date", result.getProperty(DCTerms.modified).getLiteral().toString());

        final Map<String, Resource> distributionMap = collectDistributions(result);

        checkSeehundeDistributions(distributionMap);
    }

    /**
     * The dataset itself contains two distributions, a WFS and a WMS.
     */
    @Test
    public void convert_Seehunde() throws DocumentException, IOException {

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/e95c2bed-84c0-4e7c-a194-1568887fc355.xml"));

        final Resource result = service.convert(inputDocument);

        assertTrue(result.hasLiteral(DCTerms.identifier, "e95c2bed-84c0-4e7c-a194-1568887fc355"));

        final Map<String, Resource> distributionMap = collectDistributions(result);

        assertEquals(2, distributionMap.size());
    }

    Map<String, Resource> collectDistributions(Collection<Resource> distributions) {
        final HashMap<String, Resource> result = new HashMap<>();
        for (Resource distribution : distributions) {
            final String uri = distribution.getPropertyResourceValue(DCAT.accessURL).getURI();
            result.put(uri, distribution);
        }
        return result;
    }

    /**
     * Create a Map with all distributions connected to a Dataset.
     * The accessURL of the distribution will be used as key.
     */
    Map<String, Resource> collectDistributions(Resource dataset) {
        final List<Resource> distributions = new LinkedList<>();
        StmtIterator it = dataset.listProperties(DCAT.distribution);
        while (it.hasNext()) {
            distributions.add(it.nextStatement().getResource());
        }
        return collectDistributions(distributions);
    }

    /**
     * Sometimes a geo service contains more than one distribution.
     */
    @Test
    public void convertServiceToDistributions_multipleDistributions() throws Exception {
        final Element serviceElement = saxReader.read(getClass().getResourceAsStream("/5a9a19d0-f24b-427e-9bdd-c1c7da2e84e7.xml")).getRootElement();

        final Collection<Resource> result = service.convertServiceToDistributions(serviceElement);

        assertNotNull(result);

        final Map<String, Resource> distributionMap = collectDistributions(result);

        checkSeehundeDistributions(distributionMap);
    }

    /**
     * A distribution contains an invalid URL.
     */
    @Test
    public void convert_invalidURL() throws Exception {

        service.ignoreInvalidMapping = true;

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/E3012CDB-B9EE-43BE-B5A9-1D09784F22B8.xml"));

        final Resource result = service.convert(inputDocument);

        assertNotNull(result);

        final Model model = result.getModel();
        model.write(Writer.nullWriter(), "RDF/XML-ABBREV");
    }

    /**
     * Result does not contain an <code>OnlineResource</code> with type <code>download</code>. Only a resource with
     * type <code>information</code> is given.
     */
    @Test
    public void convert_tile() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/tile.xml"));

        final Resource result = service.convert(inputDocument);

        assertTrue(result.hasLiteral(DCTerms.identifier, "e895a4b0-fdfd-4731-8137-ccefaf6e3ece"));
        assertFalse(result.hasProperty(DCAT.distribution));

        Property foafPage = ResourceFactory.createProperty("http://xmlns.com/foaf/0.1/page");

        assertTrue(result.hasProperty(foafPage));
        assertEquals("https://www.schleswig-holstein.de/DE/Landesregierung/LVERMGEOSH/Service/serviceGeobasisdaten/geodatenService_Geobasisdaten_DOP.html", result.getPropertyResourceValue(foafPage).getURI());
    }

    /**
     * Record has no temporal coverage. Therefore, the result must not contain a <code>dct:temporal</code> property.
     */
    @Test
    public void convert_missingTemporalCoverage() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/tile.xml"));

        final Resource result = service.convert(inputDocument);

        assertFalse(result.hasProperty(DCTerms.temporal));
    }

    /**
     * The <code>EX_TemporalExtent</code> does not contain usable temporal information. Therefore, no
     * <code>dct:temporal</code> property will be added.
     */
    @Test
    public void addTemporalExtend_empty() throws DocumentException {

        String xml = "<gmd:EX_TemporalExtent xmlns:gmd=\"http://www.isotc211.org/2005/gmd\">\n" +
                "  <gmd:extent>\n" +
                "    <gml:TimePeriod xmlns:gml=\"http://www.opengis.net/gml/3.2\" gml:id=\"w14182aaa\">\n" +
                "      <gml:beginPosition indeterminatePosition=\"unknown\" />\n" +
                "      <gml:endPosition indeterminatePosition=\"unknown\" />\n" +
                "    </gml:TimePeriod>\n" +
                "  </gmd:extent>\n" +
                "</gmd:EX_TemporalExtent>\n";

        Element element = new SAXReader().read(new StringReader(xml)).getRootElement();

        Resource resource = model.createResource();

        MDMetadata2Dataset.addTemporalExtend(resource, element);

        // Since there is no usable temporal information no statement must have been added.
        assertFalse(model.listStatements().hasNext());
    }

    /**
     * The <code>EX_TemporalExtent</code>  contains a date.
     */
    @Test
    public void addTemporalExtend_date() throws DocumentException {

        String xml = "<gmd:EX_TemporalExtent xmlns:gmd=\"http://www.isotc211.org/2005/gmd\" xmlns:gml=\"http://www.opengis.net/gml\">\n" +
                "<gmd:extent>\n" +
                "<gml:TimePeriod gml:id=\"timeperiod_id_a8d11de2-7468-4728-b485-cad5b58814f8\">\n" +
                "<gml:beginPosition>2020-07-10</gml:beginPosition>\n" +
                "<gml:endPosition indeterminatePosition=\"unknown\"/>\n" +
                "</gml:TimePeriod>\n" +
                "</gmd:extent>\n" +
                "</gmd:EX_TemporalExtent>";

        Element element = new SAXReader().read(new StringReader(xml)).getRootElement();

        Resource resource = model.createResource();

        MDMetadata2Dataset.addTemporalExtend(resource, element);

        // Since there is no usable temporal information no statement must have been added.
        ResIterator it = model.listResourcesWithProperty(RDF.type, DCTerms.PeriodOfTime);
        assertTrue(it.hasNext());


        /* should be:
         * <dct:PeriodOfTime>
         *   <schema:startDate rdf:datatype="http://www.w3.org/2001/XMLSchema#date">2020-07-10</schema:startDate>
         * </dct:PeriodOfTime>
         */
        Resource periodOfTime = it.next();
        assertFalse(periodOfTime.hasProperty(SchemaOrg.endDate));
        Statement stmt = periodOfTime.getProperty(SchemaOrg.startDate);
        assertNotNull(stmt);
        assertTrue(stmt.getObject().isLiteral());
        Literal startDate = stmt.getLiteral();
        assertEquals("2020-07-10", startDate.getString());
        assertEquals("http://www.w3.org/2001/XMLSchema#date", startDate.getDatatypeURI());
    }

    /**
     * There are datasets that "hide" a WMS in the info link. With an option this info link can be converted into
     * a distribution.
     */
    @Test
    public void convert_WMSinInfo() throws DocumentException, IOException {
        settings.findWMSinInfo = true;

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/c6b7ef77-dcb1-4b7a-9aac-a6d027c30a37.xml"));

        final Resource result = service.convert(inputDocument);

        final Map<String, Resource> distributionMap = collectDistributions(result);

        assertEquals(2, distributionMap.size());
        assertTrue(distributionMap.containsKey("https://sh.xplanungsplattform.de/xplan-wms/services/planwerkwms/planname/B069_0_Kernbereich-Ostseepark?request=GetCapabilities&service=WMS&version=1.3.0"));
        assertTrue(distributionMap.containsKey("https://sh.xplanungsplattform.de/xplan-wfs/services/wfs32?request=GetFeature&service=WFS&version=2.0.0&resolvedepth=*&StoredQuery_ID=urn:ogc:def:query:OGC-WFS::PlanName&planName=B069_0_Kernbereich-Ostseepark"));

        Resource distributionWMS = distributionMap.get("https://sh.xplanungsplattform.de/xplan-wms/services/planwerkwms/planname/B069_0_Kernbereich-Ostseepark?request=GetCapabilities&service=WMS&version=1.3.0");
        assertEquals(FILE_TYPE_WMS_SRC, distributionWMS.getProperty(DCTerms.format).getResource());
    }

    /**
     * Record has no temporal coverage. With the option "FIX_MISSING_TEMPORAL" the publication date will be used as
     * temporal start time.
     */
    @Test
    public void convert_addTemporalStart() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/tile.xml"));

        service.getSettings().addMissingTemporalStart = true;

        final Resource result = service.convert(inputDocument);

        assertTrue(result.hasProperty(DCTerms.temporal));

        final Resource temporal = result.getProperty(DCTerms.temporal).getObject().asResource();
        assertTrue(temporal.hasProperty(SchemaOrg.startDate));

        assertEquals("2021-08-31", temporal.getProperty(SchemaOrg.startDate).getLiteral().getLexicalForm());
        assertEquals(XSDDatatype.XSDdate, temporal.getProperty(SchemaOrg.startDate).getLiteral().getDatatype());
    }

    /**
     * Invalid URLs must not crash the transformation.
     */
    @Test
    public void convert_invalidURL2() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/E3012CDB-B9EE-43BE-B5A9-1D09784F22B8.xml"));

        final Resource result = service.convert(inputDocument);

        assertNotNull(result);

        final Model model = result.getModel();
        model.write(Writer.nullWriter(), "RDF/XML-ABBREV");
    }

    /**
     * The dataset itself contains a download link, several info links and has a couple service, too.
     */
    @Test
    public void convert_MEKUN() throws DocumentException, IOException {

        // prepare metadata for couple service
        {
            final Document document = saxReader.read(getClass().getResourceAsStream("/GetRecordsResponse_operatesOn_mekun.xml"));
            List<Element> serviceElements = document.getRootElement().element("SearchResults").elements();
            Mockito.when(cswInterface.findOperatesOn("https://registry.gdi-de.org/id/de.sh/ad15c139-cf86-4027-a46c-470bc7c52d7c")).thenReturn(serviceElements);
        }

        // prepare input element
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/GetRecordsResponse_mekun.xml"));
        final Element inputElement = inputDocument.getRootElement().element("SearchResults").elements("MD_Metadata").get(1);

        final Resource result = service.convert(inputElement);

        final Map<String, Resource> distributionMap = collectDistributions(result);

        assertEquals(2, distributionMap.size());
        assertTrue(distributionMap.containsKey("https://service.gdi-sh.de/SH_OpenGBD/feeds/Waerme/data/Waermebedarf_ad15c139-cf86-4027-a46c-470bc7c52d7c_Gemeinden.zip"));
    }

    /**
     * Test format detection in gmd:applicationProfile element (as specified in mapping table #78)
     */
    @Test
    public void convert_applicationProfile() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/applicationProfile.xml"));

        final Resource result = service.convert(inputDocument);

        final Map<String, Resource> distributionMap = collectDistributions(result);
        assertEquals(1, distributionMap.size());
        Resource distribution = distributionMap.get("https://umweltanwendungen.schleswig-holstein.de/data/meta/boden/bodkart/buek_GWS/GWS_buek250N.zip");
        assertEquals(FILE_TYPE_SHP, distribution.getProperty(DCTerms.format).getResource());
        assertEquals("Shape download", distribution.getProperty(DCTerms.title).getString());
    }

    @Test
    public void convertServiceToDistributions_2() throws DocumentException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/service2.xml"));
        final Collection<Resource> result = service.convertServiceToDistributions(inputDocument.getRootElement());

        assertNotNull(result);
        assertEquals(1, result.size());

        final Resource distribution = result.iterator().next();

        final Model model = distribution.getModel();
        assertTrue(distribution.hasProperty(RDF.type, DCAT.Distribution));
        assertTrue(distribution.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/cc-by/4.0")));
        assertTrue(distribution.hasProperty(DCTerms.format, FILE_TYPE_WMS_SRC));
        assertTrue(distribution.hasLiteral(DCATAPde.licenseAttributionByText, "© GeoBasis-DE/LVermGeo SH/CC BY 4.0"));
        assertEquals("https://service.gdi-sh.de/WMS_SH_PLANUNG?version=1.3.0&service=wms&request=GetCapabilities",
                distribution.getProperty(DCAT.accessURL).getResource().getURI());
    }

    /**
     * An invalid email address must not lead to an invalid RDF document.
     */
    @Test
    public void createFoafOrg_invalidEmail() throws DocumentException {
        final Element ciResponsibleParty = saxReader.read(getClass().getResourceAsStream("/CI_ResponsibleParty1.xml")).getRootElement();
        MDMetadata2Dataset.createFoafOrg(ciResponsibleParty, model);
        model.write(Writer.nullWriter(), "RDF/XML-ABBREV");
    }

    @Test
    public void convertServiceToDistributions_3() throws DocumentException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/service3.xml"));
        final Collection<Resource> result = service.convertServiceToDistributions(inputDocument.getRootElement());

        assertNotNull(result);
        assertEquals(1, result.size());

        final Resource distribution = result.iterator().next();

    }

    @Test
    public void convertServiceToDistributions_4() throws DocumentException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/service4.xml"));
        final Collection<Resource> result = service.convertServiceToDistributions(inputDocument.getRootElement());

        assertNotNull(result);
        assertEquals(1, result.size());

        final Resource distribution = result.iterator().next();
    }

    /**
     * Invalid URLs (only "http://") must not crash the transformation.
     */
    @Test
    public void convert_invalidURL3() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/5EA8954A-EFF2-4790-8770-2110C256BA99.xml"));

        final Resource result = service.convert(inputDocument);

        assertNotNull(result);

        final Model model = result.getModel();
        model.write(Writer.nullWriter(), "RDF/XML-ABBREV");
    }

    /**
     * Some publishers specify the licence as the URL and not the id.
     */
    @Test
    public void convert_licenceWorkaround() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/2bae1d14-9d33-4fd5-bebe-88ead246950c.xml"));
        final Resource result = service.convert(inputDocument);
        assertNotNull(result);

        assertTrue(result.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(result.hasLiteral(DCATAPde.licenseAttributionByText, "Stadt Rendsburg"));

        final Map<String, Resource> distributionMap = collectDistributions(result);
        assertEquals(1, distributionMap.size());

        final Resource distribution = distributionMap.values().iterator().next();
        assertTrue(distribution.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(distribution.hasLiteral(DCATAPde.licenseAttributionByText, "Stadt Rendsburg"));
    }

    /**
     * The format of a transfer option is specified using the gmd:applicationProfile element
     */
    @Test
    public void convert_formatAsApplicationProfile() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/107cb30a-fe45-4736-ab8c-a0a60910a4ce.xml"));
        final Resource result = service.convert(inputDocument);
        assertNotNull(result);

        final Map<String, Resource> distributionMap = collectDistributions(result);
        assertEquals(2, distributionMap.size());

        final Resource distributionZIP = distributionMap.get("https://opendatarepo.lsh.uni-kiel.de/data/elmshorn/Foerdergebiete_Soziale_Stadt.zip");
        assertTrue(distributionZIP.hasProperty(DCTerms.format, ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/SHP")));

        final Resource distributionJSON = distributionMap.get("https://opendatarepo.lsh.uni-kiel.de/data/elmshorn/Foerdergebiete_Soziale_Stadt_JSON.json");
        assertTrue(distributionJSON.hasProperty(DCTerms.format, ResourceFactory.createResource("http://publications.europa.eu/resource/authority/file-type/GEOJSON")));
    }

    /**
     * The gmd:referenceSystemInfo element is converted correctly.
     */
    @Test
    void convert_referenceSystemInfo() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/seegras.xml"));
        service.ignoreInvalidMapping = true;
        final Resource result = service.convert(inputDocument);

        assertTrue(result.hasProperty(DCTerms.conformsTo, ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/31467")));
    }

    /**
     * The gmd:referenceSystemInfo provided as a string element is converted correctly.
     */
    @Test
    void convert_referenceSystemInfo_String() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/triberg.xml"));
        service.ignoreInvalidMapping = true;
        final Resource result = service.convert(inputDocument);

        assertTrue(result.hasProperty(DCTerms.conformsTo, ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/25832")));
    }

    @Test
    public void convert_referenceSystemInfo_realExample() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/107cb30a-fe45-4736-ab8c-a0a60910a4ce.xml"));
        final Resource result = service.convert(inputDocument);
        assertNotNull(result);

        StmtIterator it = result.listProperties(DCTerms.conformsTo);
        assertTrue(it.hasNext());
        RDFNode object = it.next().getObject();

        assertFalse(object.isLiteral());
        assertEquals(ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/25832"), object);
    }

    @Test
    void addBoundingBox_null() throws DocumentException {
        Resource resource = model.createResource();
        MDMetadata2Dataset.addBoundingBox(resource, null);
        assertFalse(model.listStatements().hasNext());
    }

    @Test
    void addBoundingBox_Polygon() throws DocumentException {
        String xml = "<gmd:EX_GeographicBoundingBox xmlns:gmd=\"http://www.isotc211.org/2005/gmd\" xmlns:gco=\"http://www.isotc211.org/2005/gco\">\n" +
                " <gmd:extentTypeCode><gco:Boolean>true</gco:Boolean></gmd:extentTypeCode>\n" +
                " <gmd:westBoundLongitude><gco:Decimal>7.87</gco:Decimal></gmd:westBoundLongitude>\n" +
                " <gmd:eastBoundLongitude><gco:Decimal>11.32</gco:Decimal></gmd:eastBoundLongitude>\n" +
                " <gmd:southBoundLatitude><gco:Decimal>53.37</gco:Decimal></gmd:southBoundLatitude>\n" +
                " <gmd:northBoundLatitude><gco:Decimal>55.06</gco:Decimal></gmd:northBoundLatitude>\n" +
                "</gmd:EX_GeographicBoundingBox>";

        Element element = new SAXReader().read(new StringReader(xml)).getRootElement();
        Resource resource = model.createResource();

        MDMetadata2Dataset.addBoundingBox(resource, element);

        ResIterator it = model.listResourcesWithProperty(RDF.type, DCTerms.Location);
        assertTrue(it.hasNext());

        /* should be:
         * <dct:spatial>
         *    <dct:Location>
         *       <locn:geometry rdf:datatype="http://www.opengis.net/ont/geosparql#wktLiteral">POLYGON(( ... ))</locn:geometry>
         *    </dct:Location>
         * </dct:spatial>
         */
        Resource spatial = it.next();
        Statement stmt = spatial.getProperty(Locn.geometry);
        assertNotNull(stmt);
        assertTrue(stmt.getObject().isLiteral());
        Literal geometry = stmt.getLiteral();
        assertEquals("POLYGON((7.87 55.06,11.32 55.06,11.32 53.37,7.87 53.37,7.87 55.06))", geometry.getString());
        assertEquals("http://www.opengis.net/ont/geosparql#wktLiteral", geometry.getDatatypeURI());
    }

    /**
     * Both longitude and latitude are equal -> this is a point.
     *
     * @throws DocumentException
     */
    @Test
    void addBoundingBox_Point() throws DocumentException {
        String xml = "<gmd:EX_GeographicBoundingBox xmlns:gmd=\"http://www.isotc211.org/2005/gmd\" xmlns:gco=\"http://www.isotc211.org/2005/gco\">\n" +
                " <gmd:extentTypeCode><gco:Boolean>true</gco:Boolean></gmd:extentTypeCode>\n" +
                " <gmd:westBoundLongitude><gco:Decimal>7.87</gco:Decimal></gmd:westBoundLongitude>\n" +
                " <gmd:eastBoundLongitude><gco:Decimal>7.87</gco:Decimal></gmd:eastBoundLongitude>\n" +
                " <gmd:southBoundLatitude><gco:Decimal>53.37</gco:Decimal></gmd:southBoundLatitude>\n" +
                " <gmd:northBoundLatitude><gco:Decimal>53.37</gco:Decimal></gmd:northBoundLatitude>\n" +
                "</gmd:EX_GeographicBoundingBox>";

        Element element = new SAXReader().read(new StringReader(xml)).getRootElement();
        Resource resource = model.createResource();

        MDMetadata2Dataset.addBoundingBox(resource, element);

        ResIterator it = model.listResourcesWithProperty(RDF.type, DCTerms.Location);
        assertTrue(it.hasNext());

        /* should be:
         * <dct:spatial>
         *    <dct:Location>
         *       <locn:geometry rdf:datatype="http://www.opengis.net/ont/geosparql#wktLiteral">POINT( ... )</locn:geometry>
         *    </dct:Location>
         * </dct:spatial>
         */
        Resource spatial = it.next();
        Statement stmt = spatial.getProperty(Locn.geometry);
        assertNotNull(stmt);
        assertTrue(stmt.getObject().isLiteral());
        Literal geometry = stmt.getLiteral();
        assertEquals("POINT(7.87 53.37)", geometry.getString());
        assertEquals("http://www.opengis.net/ont/geosparql#wktLiteral", geometry.getDatatypeURI());
    }

    /**
     * Latitude is equal -> this is a line.
     */
    @Test
    void addBoundingBox_Line1() throws DocumentException {
        String xml = "<gmd:EX_GeographicBoundingBox xmlns:gmd=\"http://www.isotc211.org/2005/gmd\" xmlns:gco=\"http://www.isotc211.org/2005/gco\">\n" +
                " <gmd:extentTypeCode><gco:Boolean>true</gco:Boolean></gmd:extentTypeCode>\n" +
                " <gmd:westBoundLongitude><gco:Decimal>9.68</gco:Decimal></gmd:westBoundLongitude>\n" +
                " <gmd:eastBoundLongitude><gco:Decimal>9.69</gco:Decimal></gmd:eastBoundLongitude>\n" +
                " <gmd:southBoundLatitude><gco:Decimal>53.76</gco:Decimal></gmd:southBoundLatitude>\n" +
                " <gmd:northBoundLatitude><gco:Decimal>53.76</gco:Decimal></gmd:northBoundLatitude>\n" +
                "</gmd:EX_GeographicBoundingBox>";

        Element element = new SAXReader().read(new StringReader(xml)).getRootElement();
        Resource resource = model.createResource();

        MDMetadata2Dataset.addBoundingBox(resource, element);

        ResIterator it = model.listResourcesWithProperty(RDF.type, DCTerms.Location);
        assertTrue(it.hasNext());

        /* should be:
         * <dct:spatial>
         *    <dct:Location>
         *       <locn:geometry rdf:datatype="http://www.opengis.net/ont/geosparql#wktLiteral">LINESTRING( ... )</locn:geometry>
         *    </dct:Location>
         * </dct:spatial>
         */
        Resource spatial = it.next();
        Statement stmt = spatial.getProperty(Locn.geometry);
        assertNotNull(stmt);
        assertTrue(stmt.getObject().isLiteral());
        Literal geometry = stmt.getLiteral();
        assertEquals("LINESTRING(9.68 53.76,9.69 53.76)", geometry.getString());
        assertEquals("http://www.opengis.net/ont/geosparql#wktLiteral", geometry.getDatatypeURI());
    }

    /**
     * Longitude is equal -> this is a line.
     */
    @Test
    void addBoundingBox_Line2() throws DocumentException {
        String xml = "<gmd:EX_GeographicBoundingBox xmlns:gmd=\"http://www.isotc211.org/2005/gmd\" xmlns:gco=\"http://www.isotc211.org/2005/gco\">\n" +
                " <gmd:extentTypeCode><gco:Boolean>true</gco:Boolean></gmd:extentTypeCode>\n" +
                " <gmd:westBoundLongitude><gco:Decimal>9.68</gco:Decimal></gmd:westBoundLongitude>\n" +
                " <gmd:eastBoundLongitude><gco:Decimal>9.68</gco:Decimal></gmd:eastBoundLongitude>\n" +
                " <gmd:southBoundLatitude><gco:Decimal>53.76</gco:Decimal></gmd:southBoundLatitude>\n" +
                " <gmd:northBoundLatitude><gco:Decimal>53.77</gco:Decimal></gmd:northBoundLatitude>\n" +
                "</gmd:EX_GeographicBoundingBox>";

        Element element = new SAXReader().read(new StringReader(xml)).getRootElement();
        Resource resource = model.createResource();

        MDMetadata2Dataset.addBoundingBox(resource, element);

        ResIterator it = model.listResourcesWithProperty(RDF.type, DCTerms.Location);
        assertTrue(it.hasNext());

        /* should be:
         * <dct:spatial>
         *    <dct:Location>
         *       <locn:geometry rdf:datatype="http://www.opengis.net/ont/geosparql#wktLiteral">LINESTRING( ... )</locn:geometry>
         *    </dct:Location>
         * </dct:spatial>
         */
        Resource spatial = it.next();
        Statement stmt = spatial.getProperty(Locn.geometry);
        assertNotNull(stmt);
        assertTrue(stmt.getObject().isLiteral());
        Literal geometry = stmt.getLiteral();
        assertEquals("LINESTRING(9.68 53.76,9.68 53.77)", geometry.getString());
        assertEquals("http://www.opengis.net/ont/geosparql#wktLiteral", geometry.getDatatypeURI());
    }

    @Test
    public void convert_referenceSystemInfo_realExample2() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/7597ab47-62fb-4d10-8ec2-fc80f5296309.xml"));
        final Resource result = service.convert(inputDocument);
        assertNotNull(result);

        List<RDFNode> referenceSystems = new ArrayList<>();
        StmtIterator it = result.listProperties(DCTerms.conformsTo);
        while (it.hasNext()) {
            RDFNode object = it.next().getObject();
            referenceSystems.add(object);
        }

        assertEquals(3, referenceSystems.size());

        assertFalse(referenceSystems.get(0).isLiteral());
        assertFalse(referenceSystems.get(1).isLiteral());
        assertFalse(referenceSystems.get(2).isLiteral());
        assertTrue(referenceSystems.contains(ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/31467")));
        assertTrue(referenceSystems.contains(ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/31468")));
        assertTrue(referenceSystems.contains(ResourceFactory.createResource("http://www.opengis.net/def/crs/EPSG/0/4647")));
    }

    /**
     * This dataset does not have gmd:topicCategory elements.
     */
    @Test
    public void convert_missingTopicCategories() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/1f16fa2b-7bc7-4cb4-a94c-283355c10b1e.xml"));
        final Resource result = service.convert(inputDocument);
        assertNotNull(result);

        assertNotNull(result.getProperty(DCAT.theme));
    }

    @Test
    public void collectResponsibleParties() throws Exception {
        final Element input = saxReader.read(getClass().getResourceAsStream("/seegras.xml")).getRootElement();
        final Map<String, Element> result = MDMetadata2Dataset.collectResponsibleParties(input);

        assertEquals(3, result.size());
        assertTrue(result.containsKey("publisher"));
        assertTrue(result.containsKey("pointOfContact"));
        assertTrue(result.containsKey("originator"));
    }

    /**
     * This dataset does not have a gmd:pointOfContact element. Therefore, the gmd:contact must be used instead.
     *
     * @throws Exception
     */
    @Test
    public void collectResponsibleParties_noPointOfContact() throws Exception {
        final Element input = saxReader.read(getClass().getResourceAsStream("/7b510ce5-d4d5-48d0-867b-c80778cf453c.xml")).getRootElement();

        final Map<String, Element> result = MDMetadata2Dataset.collectResponsibleParties(input);

        assertEquals(1, result.size());
        assertTrue(result.containsKey("pointOfContact"));
    }

    /**
     * Parameter is not a gmd:MD_Metadata element.
     */
    @Test
    public void convertServiceToDistributions_invalidElement() {
        Document document = DocumentHelper.createDocument();
        Element element = document.addElement("test");

        assertThrows(IllegalArgumentException.class, () -> service.convertServiceToDistributions(element));
    }

    @Test
    public void convertServiceToDistributions_inputIsNoService() {
        Document document = DocumentHelper.createDocument();
        Element mdMetadata = document.addElement(QName.get("MD_Metadata", NS_GMD));

        mdMetadata.addElement(QName.get("hierarchyLevel", NS_GMD))
                .addElement(QName.get("MD_ScopeCode", NS_GMD))
                .addAttribute("codeListValue", "application");

        assertThrows(IllegalArgumentException.class, () -> service.convertServiceToDistributions(mdMetadata));
    }

    /**
     * This 'service' does not have a usable URI.
     */
    @Test
    public void convertServiceToDistributions_noDistribution() {
        Document document = DocumentHelper.createDocument();
        Element mdMetadata = document.addElement(QName.get("MD_Metadata", NS_GMD));

        // gmd:hierarchyLevel/gmd:MD_ScopeCode[@codeListValue='service']
        mdMetadata.addElement(QName.get("hierarchyLevel", NS_GMD))
                .addElement(QName.get("MD_ScopeCode", NS_GMD))
                .addAttribute("codeListValue", "service");

        Collection<Resource> result = service.convertServiceToDistributions(mdMetadata);
        assertNotNull(result);
        assertTrue(result.isEmpty());
    }

    /**
     * The gmd prefix is not bound in this document.
     */
    @Test
    public void convertServiceToDistributions_noGmdPrefix() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/d65a2641-1a2e-4011-b729-2e5578efe68c.xml"));
        Collection<Resource> result = service.convertServiceToDistributions(inputDocument.getRootElement());
        assertNotNull(result);
        assertEquals(1, result.size());
    }

    /**
     * Parameter is not a gmd:MD_Metadata element.
     */
    @Test
    public void convert_invalidElement() {
        Document document = DocumentHelper.createDocument();
        Element element = document.addElement("test");

        assertThrows(IllegalArgumentException.class, () -> service.convert(element));
    }

    /**
     * Parameter is not a EX_GeographicBoundingBox element.
     */
    @Test
    public void addBoundingBox_invalidElement() {
        Document document = DocumentHelper.createDocument();
        Element element = document.addElement("test");

        assertThrows(IllegalArgumentException.class, () -> MDMetadata2Dataset.addBoundingBox(ResourceFactory.createResource(), element));
    }

    /**
     * The gmd:EX_TemporalExtent does not have a gmd:extent child element
     */
    @Test
    public void addTemporalExtend_missingChildElement() {
        Document document = DocumentHelper.createDocument();
        Element element = document.addElement(QName.get("EX_TemporalExtent", NS_GMD));

        assertThrows(IllegalArgumentException.class, () -> MDMetadata2Dataset.addTemporalExtend(ResourceFactory.createResource(), element));
    }

    /**
     * Some information in gmd:DQ_DomainConsistency will be mapped into dct:conformTo properties.
     */
    @Test
    public void convert_DomainConsistency() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/triberg.xml"));
        service.ignoreInvalidMapping = true;
        final Resource result = service.convert(inputDocument);

        Set<String> conformsTo = new HashSet<>();
        StmtIterator it = result.listProperties(DCTerms.conformsTo);
        while (it.hasNext()) {
            conformsTo.add(it.next().getResource().getURI());
        }
        assertTrue(conformsTo.contains("http://data.europa.eu/eli/reg/2010/1089/oj"));
    }

    /**
     * Datasets with the keyword "inspireidentifiziert" get a property dcatap:applicableLegislation
     */
    @Test
    public void convert_inspire() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/e95c2bed-84c0-4e7c-a194-1568887fc355.xml"));
        final Resource result = service.convert(inputDocument);

        assertTrue(collectApplicableLegislation(result).contains("http://data.europa.eu/eli/dir/2007/2/2019-06-26"));
    }

    private Set<String> collectApplicableLegislation(Resource result) {
        Set<String> conformsTo = new HashSet<>();
        StmtIterator it = result.listProperties(DCATAP.applicableLegislation);
        while (it.hasNext()) {
            conformsTo.add(it.next().getResource().getURI());
        }
        return conformsTo;
    }

    /**
     * The MD_Distribution element of this dataset does not have a distributionFormat element.
     */
    @Test
    public void convert_missingFormat() throws Exception {
        final String excpetedExceptionMessage = "Could not determine format of distribution https://gis.schleswig-flensburg.de/arcgis/services/WMS_SL/LSG_SL/MapServer/WFSServer";
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/af6da150-8be3-4f84-b34f-bae1b545e52d.xml"));
        RuntimeException ex = assertThrows(RuntimeException.class, () -> service.convert(inputDocument));
        assertEquals(excpetedExceptionMessage, ex.getMessage());
    }

    /**
     * This entry is a service marked as "dataset". Therefore, it has a srv:SV_ServiceIdentification instead of a
     * gmd:MD_DataIdentification element.
     */
    @Test
    public void convert_serviceIdentification() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/f552801e-2eea-4c55-9cf6-e7e3deb13d40.xml"));
        final Resource result = service.convert(inputDocument);

        assertNotNull(result);

        assertTrue(result.hasProperty(DCTerms.title), "dataset has a title property");
    }

    @Test
    public void convertServiceToDistributions_atomfeed() throws Exception {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/ec98945c-4a92-4ab3-ad20-fd8912494546.xml"));
        final Collection<Resource> result = service.convertServiceToDistributions(inputDocument.getRootElement());
        assertNotNull(result);
        assertEquals(2, result.size());

        int distributionATOM = 0;
        int distributionHTML = 0;
        for (Resource distribution : result) {
            if (distribution.hasProperty(DCTerms.format, FILE_TYPE_ATOM)) {
                distributionATOM++;
                assertEquals("https://service.gdi-sh.de/SH_OpenGBD/feeds/DOP20/DOP20.xml",
                        distribution.getProperty(DCAT.accessURL).getResource().getURI());
            } else if (distribution.hasProperty(DCTerms.format, FILE_TYPE_HTML)) {
                distributionHTML++;
                assertEquals("https://service.gdi-sh.de/AtomFeedViewer?feed=https://service.gdi-sh.de/SH_OpenGBD/feeds/DOP20/DOP20.xml",
                        distribution.getProperty(DCAT.accessURL).getResource().getURI());
            }
            assertTrue(distribution.hasProperty(RDF.type, DCAT.Distribution));
            assertTrue(distribution.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/cc-by/4.0")));
            assertTrue(distribution.hasLiteral(DCATAPde.licenseAttributionByText, "© GeoBasis-DE/LVermGeo SH/CC BY 4.0"));
        }

        assertEquals(1, distributionHTML, "one HTML distribution (the Atom feed viewer)");
        assertEquals(1, distributionATOM, "one ATOM distribution (the feed itself)");

    }

    /**
     * The XPlanungsplattform sometimes sends incorrect license information for some datasets. Although the id of the
     * license is invalide the URI of the license is correct.
     */
    @Test
    public void convert_incorrectLicense() throws DocumentException, IOException {
        settings.findWMSinInfo = true;

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/e936ff9d-ba13-41eb-b0fa-41d5124496e0.xml"));
        final Resource dataset = service.convert(inputDocument);

        assertNotNull(dataset);
        assertTrue(dataset.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(dataset.hasLiteral(DCATAPde.licenseAttributionByText, "Stadt Flensburg"));

        final Resource distribution = collectDistributions(dataset).values().iterator().next();
        assertTrue(distribution.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(distribution.hasLiteral(DCATAPde.licenseAttributionByText, "Stadt Flensburg"));
    }

    /**
     * This serivce for an ATOM feed contains two transfer options: one for the feed itself and one for a viewer
     * application.
     */
    @Test
    public void convertServiceToDistributions_atomfeed2() throws Exception {
        final String expectedFeedURL = "https://dienste.gdi-sh.de/SH_OpenGBD/feeds/Atom_SH_ALKIS_vereinf_OpenGBD/Atom_SH_ALKIS_vereinf_OpenGBD.xml";
        final String expectedViewerURL = "https://service.gdi-sh.de/AtomFeedViewer?feed=https://service.gdi-sh.de/SH_OpenGBD/feeds/Atom_SH_ALKIS_vereinf_OpenGBD/Atom_SH_ALKIS_vereinf_OpenGBD.xml";

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/9066e59a-c4e5-4254-b3e1-eb4eb16bdc53.xml"));
        final Collection<Resource> result = service.convertServiceToDistributions(inputDocument.getRootElement());
        assertNotNull(result);
        assertEquals(2, result.size());

        int distributionATOM = 0;
        int distributionHTML = 0;
        for (Resource distribution : result) {
            if (distribution.hasProperty(DCTerms.format, FILE_TYPE_ATOM)) {
                distributionATOM++;
                assertEquals(expectedFeedURL, distribution.getProperty(DCAT.accessURL).getResource().getURI());
            } else if (distribution.hasProperty(DCTerms.format, FILE_TYPE_HTML)) {
                distributionHTML++;
                assertEquals(expectedViewerURL, distribution.getProperty(DCAT.accessURL).getResource().getURI());
            }
            assertTrue(distribution.hasProperty(RDF.type, DCAT.Distribution));
            assertTrue(distribution.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/cc-by/4.0")));
            assertTrue(distribution.hasLiteral(DCATAPde.licenseAttributionByText, "© GeoBasis-DE/LVermGeo SH/CC BY 4.0"));
        }

        assertEquals(1, distributionHTML, "one HTML distribution (the Atom feed viewer)");
        assertEquals(1, distributionATOM, "one ATOM distribution (the feed itself)");
    }

    /**
     * An incorrect code list specification (in this case for MD_MaintenanceFrequencyCode) must not stop the process.
     */
    @Test
    public void convert_incorrectMaintenanceFrequencyCode() throws DocumentException, IOException {
        settings.findWMSinInfo = true;

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/bb585473-0987-436a-bc7a-132880a4df0d.xml"));
        final Resource dataset = service.convert(inputDocument);

        assertNotNull(dataset);
        assertEquals("bb585473-0987-436a-bc7a-132880a4df0d", dataset.getProperty(DCTerms.identifier).getString());
    }

    /**
     * Detect multiple formats (SHP and GML) in one dataset.
     * The WMS link is so broken that it won't be detected as a WMS.
     */
    @Test
    public void convert_multiFormat() throws DocumentException, IOException {
        settings.findWMSinInfo = true;

        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/6782583C-9747-4C16-B94D-3EAF2BFDB171.xml"));

        final Resource result = service.convert(inputDocument);

        final Map<String, Resource> distributionMap = collectDistributions(result);
        assertEquals(2, distributionMap.size());

        Resource distribution = distributionMap.get("https://opendata.schleswig-holstein.de/dataset/d501a7ab-a85e-4988-a98e-8f3ae3ab5df2/resource/1bcad7cd-9c94-4a06-bbbc-c7929bb37751/download/lsg.zip");
        assertEquals(FILE_TYPE_SHP, distribution.getProperty(DCTerms.format).getResource());
        assertEquals("Landschaftsschutzgebiete (1:25.000)", distribution.getProperty(DCTerms.title).getString());

        Resource distribution2 = distributionMap.get("https://umweltgeodienste.schleswig-holstein.de/WFS_UWAT?service=wfs&version=2.0.0&request=GetFeature&typeNames=app:lsg");
        assertEquals(FILE_TYPE_GML, distribution2.getProperty(DCTerms.format).getResource());
        assertEquals("WFS Landschaftsschutzgebiete (1:25.000)", distribution2.getProperty(DCTerms.title).getString());
    }

    /**
     * Specification of the HVD category as free text in characterString elements
     * (in accordance with the GDI-DE mapping rules of 2024-03-20 variant a)
     */
    @Test
    public void convert_highValueDataset_a() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/hvd1.xml"));
        final Resource result = service.convert(inputDocument);

        assertEquals("http://data.europa.eu/eli/reg_impl/2023/138/oj",
                result.getPropertyResourceValue(DCATAP.applicableLegislation).getURI());
        assertEquals("http://data.europa.eu/bna/c_ac64a52d",
                result.getPropertyResourceValue(DCATAP.hvdCategory).getURI());
    }

    /**
     * Referencing the EU vocabulary for HVD categories as gmx:Anchor elements
     * (in accordance with the GDI-DE mapping rules of 2024-03-20 variant b)
     */
    @Test
    public void convert_highValueDataset_b() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/hvd2.xml"));
        final Resource result = service.convert(inputDocument);

        assertEquals("http://data.europa.eu/eli/reg_impl/2023/138/oj",
                result.getPropertyResourceValue(DCATAP.applicableLegislation).getURI());
        assertEquals("http://data.europa.eu/bna/c_ac64a52d",
                result.getPropertyResourceValue(DCATAP.hvdCategory).getURI());
    }

    /**
     * In some source systems, German and English titles are strangely combined in one text.
     */
    @Test
    public void convert_mixedGermanEnglishTitles() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/00729c12-e689-446d-83bd-dd8cfd0eeeb8.xml"));
        final Resource result = service.convert(inputDocument);

        assertTrue(result.hasProperty(DCTerms.title, ResourceFactory.createLangLiteral(
                "Robben im Schleswig-Holsteinischen Wattenmeer 1995", "de")));
        assertTrue(result.hasProperty(DCTerms.title, ResourceFactory.createLangLiteral(
                "Seals in the Schleswig-Holstein Wadden Sea in 1995", "en")));
    }

    /**
     * This WFS has a different output format syntax for shape files.
     */
    @Test
    public void convert_wfsWithShape() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/710b07d4-2b80-4e81-bdb6-c60f06ffbb9b.xml"));
        final Resource result = service.convert(inputDocument);

        final Map<String, Resource> distributionMap = collectDistributions(result);
        assertEquals(1, distributionMap.size());

        Resource distribution = distributionMap.get(distributionMap.keySet().iterator().next()); // first and only entry
        assertEquals(FILE_TYPE_SHP, distribution.getProperty(DCTerms.format).getResource());
    }

    /**
     * This geo series does not have a download option (which is totally valid). If the configuration "skip
     * series without distribution" is set this series will be ignored.
     */
    @Test
    public void convert_SeriesWithoutDistribution() throws DocumentException, IOException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/d3c6c074-4d69-4b1f-9efb-dbd9644f82c0.xml"));

        service.getSettings().ignoreSeriesWithoutDistribution = true;

        final Resource result = service.convert(inputDocument);

        assertNull(result);
    }

    @Test
    public void createDataService() throws DocumentException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/service1.xml"));

        final Resource result = service.createDataService(inputDocument.getRootElement(), ModelFactory.createDefaultModel());

        assertNotNull(result);
        assertTrue(result.hasProperty(RDF.type, DCAT.DataService));
        assertEquals("Preußische Landesaufnahme (Chronologen, Darstellungsdienst, Schleswig-Holstein) - kostenfrei",
                result.getRequiredProperty(DCTerms.title).getLiteral().getString());
        assertEquals("Dieser Darstellungsdienst stellt die Chronologen der Preußischen Landesaufnahme dar. Die historischen Karten sind kulturhistorische Zeitzeugen der Kartographie und spiegeln im Besonderen die politische Machtverteilung wider. Die Preußische Landesaufnahme wurde in den Jahren 1878 bis 1880 aufgenommen und zeigt in den einzelnen Messtischblättern die zu dieser Zeit vorherrschende Topographie. Der Kartenmaßstab der einzelnen Messtischblätter ist 1:25.000 und deckt sich im Blattschnitt mit denen der Topographischen Karte 1:25.000 (TK25). Zusätzlich sind unter dem Begriff \"Chronologen\" Kartenblätter erhältlich, die über die Erstausgabe 1878/80 hinaus als Reproduktionen weitere Ausgaben des Topographischen Kartenwerks 1:25.000 (TK25) bis 1950 beinhalten.",
                result.getRequiredProperty(DCTerms.description).getLiteral().getString());

        assertTrue(result.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/dl-by-de/2.0")));
        assertTrue(result.hasLiteral(DCATAPde.licenseAttributionByText, "© GDI-SH"));
        assertEquals("2022-05-31", result.getProperty(DCTerms.modified).getLiteral().getString());
        assertEquals("https://service.gdi-sh.de/WMS_SH_FD_Chronologen_KF?Service=wms&version=1.3.0&request=GetCapabilities",
                result.getProperty(DCAT.endpointDescription).getResource().getURI());
        assertEquals("https://service.gdi-sh.de/WMS_SH_FD_Chronologen_KF",
                result.getProperty(DCAT.endpointURL).getResource().getURI());
        assertTrue(result.hasProperty(DCAT.servesDataset, ResourceFactory.createResource("https://registry.gdi-de.org/id/de.sh/7164d86e-8c55-4e08-ab1f-601b3002d1d1")));
    }

    @Test
    public void createDataService_withHvdMark() throws DocumentException {
        final Document inputDocument = saxReader.read(getClass().getResourceAsStream("/64d6fe33-407e-438c-b11d-122cdb7ff63a.xml"));

        final Resource result = service.createDataService(inputDocument.getRootElement(), ModelFactory.createDefaultModel());

        assertNotNull(result);
        assertTrue(result.hasProperty(RDF.type, DCAT.DataService));
        assertEquals("INSPIRE WMS SH Existierende Bodennutzung ALKIS (Darstellungsdienst)",
                result.getRequiredProperty(DCTerms.title).getLiteral().getString());
        assertEquals("Dieser Dienst stellt für das INSPIRE-Thema Bodennutzung (Existierende Bodennutzung) aus ALKIS umgesetzte Daten bereit.",
                result.getRequiredProperty(DCTerms.description).getLiteral().getString());

        assertTrue(result.hasProperty(DCTerms.license, ResourceFactory.createResource("http://dcat-ap.de/def/licenses/cc-by/4.0")));
        assertTrue(result.hasLiteral(DCATAPde.licenseAttributionByText, "© GeoBasis-DE/LVermGeo SH/CC BY 4.0"));
        assertEquals("2024-12-10", result.getProperty(DCTerms.modified).getLiteral().getString());
        assertEquals("https://service.gdi-sh.de/SH_INSPIREVIEW_AIII_ELU?service=WMS&version=1.3.0&request=GetCapabilities",
                result.getProperty(DCAT.endpointDescription).getResource().getURI());
        assertEquals("https://service.gdi-sh.de/SH_INSPIREVIEW_AIII_ELU",
                result.getProperty(DCAT.endpointURL).getResource().getURI());
        assertTrue(result.hasProperty(DCAT.servesDataset, ResourceFactory.createResource("https://registry.gdi-de.org/id/de.sh/f5626e31-7674-446b-be64-af93d494cd0b")));

        assertTrue(collectApplicableLegislation(result).contains("http://data.europa.eu/eli/reg_impl/2023/138/oj"));
        assertEquals("http://data.europa.eu/bna/c_ac64a52d",
                result.getPropertyResourceValue(DCATAP.hvdCategory).getURI());
    }

    @Test
    public void getServiceEndpoint_norderstedtWFS() throws DocumentException {
        final Element inputDocument = saxReader.read(getClass().getResourceAsStream("/norderstedt_wfs.xml")).getRootElement();
        String result = MDMetadata2Dataset.getServiceEndpoint(inputDocument);
        assertEquals("https://geoservice.norderstedt.de/geoserver/pub/wfs?SERVICE=WFS&version=2.0.0&request=GetCapabilities", result);
    }

    @Test
    public void createDataService_norderstedtWMS() throws DocumentException {
        final Element input = saxReader.read(getClass().getResourceAsStream("/norderstedt_wms.xml")).getRootElement();
        Resource dataService = service.createDataService(input, model);
        assertEquals("https://geoservice.norderstedt.de/geoserver/pub/soz_Wohlfahrt/wms", dataService.getPropertyResourceValue(DCAT.endpointURL).getURI());
    }

    @Test
    public void convertServiceToDistributions_norderstedtWMS() throws DocumentException {
        final Element input = saxReader.read(getClass().getResourceAsStream("/norderstedt_wms.xml")).getRootElement();
        Collection<Resource> distributions = service.convertServiceToDistributions(input);
        assertEquals(1, distributions.size());
        Resource distribution = distributions.iterator().next();

        assertEquals("https://geoservice.norderstedt.de/geoserver/pub/soz_Wohlfahrt/wms?SERVICE=WMS&version=1.3.0&request=GetCapabilities", distribution.getPropertyResourceValue(DCAT.accessURL).getURI());
    }

    @Test
    public void getServiceEndpoint_norderstedtWMS() throws DocumentException {
        final Element inputDocument = saxReader.read(getClass().getResourceAsStream("/norderstedt_wms.xml")).getRootElement();
        String result = MDMetadata2Dataset.getServiceEndpoint(inputDocument);
        assertEquals("https://geoservice.norderstedt.de/geoserver/pub/soz_Wohlfahrt/wms?SERVICE=WMS&version=1.3.0&request=GetCapabilities", result);
    }

    @Test
    public void convert_norderstedtDataset() throws DocumentException, IOException {
        final Element wmsElement = saxReader.read(getClass().getResourceAsStream("/norderstedt_wms.xml")).getRootElement();
        final Element wfsElement = saxReader.read(getClass().getResourceAsStream("/norderstedt_wfs.xml")).getRootElement();
        final Element input = saxReader.read(getClass().getResourceAsStream("/norderstedt_dataset.xml")).getRootElement();

        Mockito.when(cswInterface.findOperatesOn("https://registry.gdi-de.org/id/de.sh/ea7599f9-0134-4a3f-ab46-5116addc1bb3")).thenReturn(Arrays.asList(wfsElement, wmsElement));

        Resource dataset = service.convert(input);

        final Map<String, Resource> distributionMap = collectDistributions(dataset);
        assertEquals(4, distributionMap.size());

        Resource distributionWMS = distributionMap.get("https://geoservice.norderstedt.de/geoserver/pub/soz_Wohlfahrt/wms?SERVICE=WMS&version=1.3.0&request=GetCapabilities");
        Resource distributionWFS = distributionMap.get("https://geoservice.norderstedt.de/geoserver/pub/wfs?SERVICE=WFS&version=2.0.0&request=GetCapabilities");
        Resource distributionCSV = distributionMap.get("https://geoservice.norderstedt.de/geoserver/pub/ows?service=WFS&version=1.0.0&request=GetFeature&typeName=pub:soz_Wohlfahrt&outputFormat=CSV");
        Resource distributionSHP = distributionMap.get("https://geoservice.norderstedt.de/geoserver/pub/ows?service=WFS&version=1.0.0&request=GetFeature&typeName=pub:soz_Wohlfahrt&outputFormat=SHAPE-ZIP");

        assertNotNull(distributionCSV);
        assertNotNull(distributionSHP);
        assertNotNull(distributionWFS);
        assertNotNull(distributionWMS);

        assertEquals("http://publications.europa.eu/resource/authority/file-type/WFS_SRVC",
                distributionWFS.getPropertyResourceValue(DCTerms.format).getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/WMS_SRVC",
                distributionWMS.getPropertyResourceValue(DCTerms.format).getURI());
    }

    @Test
    public void convert_variousFormats() throws DocumentException, IOException {

        final Element input = saxReader.read(getClass().getResourceAsStream("/1c82089a-313e-4c25-9389-0b704c885401.xml")).getRootElement();

        Resource dataset = service.convert(input);

        final Map<String, Resource> distributionMap = collectDistributions(dataset);
        assertEquals(3, distributionMap.size());

        Resource distributionPDF = distributionMap.get("https://www.schleswig-holstein.de/mm/downloads/LFU/Geologie/GUEK250_mit_erlaeuterung.pdf");
        Resource distributionSHP = distributionMap.get("https://umweltanwendungen.schleswig-holstein.de/data/meta/boden/geokart/guek250/GUEK250_abgabe.zip");
        Resource distributionJPG = distributionMap.get("https://umweltanwendungen.schleswig-holstein.de/data/meta/boden/geokart/guek250/guek250.jpg");

        assertNotNull(distributionJPG);
        assertNotNull(distributionSHP);
        assertNotNull(distributionPDF);

        assertEquals("http://publications.europa.eu/resource/authority/file-type/SHP",
                distributionSHP.getPropertyResourceValue(DCTerms.format).getURI());
        assertEquals("http://publications.europa.eu/resource/authority/file-type/PDF",
                distributionPDF.getPropertyResourceValue(DCTerms.format).getURI());
    }

    /**
     * This is a read dataset series with the correct MD_ScopeCode
     */
    @Test
    public void isDatasetSeries_realSeries() throws DocumentException {
        final Element input = saxReader.read(getClass().getResourceAsStream("/f7f90143-c2ad-46b2-934d-93dfd5e0f031.xml")).getRootElement();
        MDMetadata2Dataset.addNamespaces(input);
        boolean result = MDMetadata2Dataset.isDatasetSeries(input);
        assertTrue(result);
    }

    /**
     * This dataset series has been incorrectly assigned the MD_ScopeCode dataset. But the title contains "(Serie)".
     */
    @Test
    public void isDatasetSeries_title() throws DocumentException {
        final Element input = saxReader.read(getClass().getResourceAsStream("/7b510ce5-d4d5-48d0-867b-c80778cf453c.xml")).getRootElement();
        MDMetadata2Dataset.addNamespaces(input);
        boolean result = MDMetadata2Dataset.isDatasetSeries(input);
        assertTrue(result);
    }

    /**
     * This is a dataset and not a dataset series.
     */
    @Test
    public void isDatasetSeries_dataset() throws DocumentException {
        final Element input = saxReader.read(getClass().getResourceAsStream("/1c82089a-313e-4c25-9389-0b704c885401.xml")).getRootElement();
        MDMetadata2Dataset.addNamespaces(input);
        boolean result = MDMetadata2Dataset.isDatasetSeries(input);
        assertFalse(result, "This is not a dataset series.");
    }
}
