package de.landsh.opendata.csw2dcat;

import io.undertow.io.Sender;
import io.undertow.server.HttpServerExchange;
import io.undertow.util.HeaderMap;
import io.undertow.util.Headers;
import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mockito;
import org.mockserver.client.MockServerClient;
import org.mockserver.integration.ClientAndServer;
import org.mockserver.model.Header;
import org.mockserver.model.HttpStatusCode;
import uk.org.webcompere.systemstubs.environment.EnvironmentVariables;
import uk.org.webcompere.systemstubs.jupiter.SystemStub;
import uk.org.webcompere.systemstubs.jupiter.SystemStubsExtension;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockserver.integration.ClientAndServer.startClientAndServer;
import static org.mockserver.matchers.Times.exactly;
import static org.mockserver.model.HttpRequest.request;
import static org.mockserver.model.HttpResponse.response;

@ExtendWith(SystemStubsExtension.class)
public class ServerTests {
    private static ClientAndServer mockServer;
    @SystemStub
    private EnvironmentVariables environmentVariables;

    @BeforeAll
    public static void startServer() {
        mockServer = startClientAndServer(1080);
    }

    @AfterAll
    public static void stopServer() {
        mockServer.stop(true);
        while (!mockServer.hasStopped(3, 100L, TimeUnit.MILLISECONDS)) {
        }
    }

    @BeforeEach
    public void setUp() {
        Server.bridgeSettings.clear();
    }

    @Test
    public void readEnvironmentVariables_single() {
        environmentVariables.set("BRIDGE_URL", "http://localhost:8081/csw");
        environmentVariables.set("BRIDGE_FIND_WMS_IN_INFO", "true");
        environmentVariables.set("BRIDGE_TYPE_EXCLUDE", "application,tile");

        Server.readEnvironmentVariables();

        assertEquals(1, Server.bridgeSettings.size());
        assertEquals("http://localhost:8081/csw", Server.bridgeSettings.get(0).cswURL);
        assertTrue(Server.bridgeSettings.get(0).findWMSinInfo);
        assertEquals("[application, tile]", Arrays.toString(Server.bridgeSettings.get(0).typeExclude));
    }

    @Test
    public void readEnvironmentVariables_removeKeyword() {
        environmentVariables.set("BRIDGE_URL", "http://localhost:8081/csw");
        environmentVariables.set("BRIDGE_REMOVE_KEYWORDS", "opendata");
        Server.readEnvironmentVariables();

        assertEquals(1, Server.bridgeSettings.size());
        assertNotNull(Server.bridgeSettings.get(0).keywordRemove);
        assertEquals(1, Server.bridgeSettings.get(0).keywordRemove.size());
        assertTrue(Server.bridgeSettings.get(0).keywordRemove.contains("opendata"));
    }

    @Test
    public void readEnvironmentVariables_removeMultipleKeyword() {
        environmentVariables.set("BRIDGE_URL", "http://localhost:8081/csw");
        environmentVariables.set("BRIDGE_REMOVE_KEYWORDS", "opendata,foo,bar");
        Server.readEnvironmentVariables();

        assertEquals(1, Server.bridgeSettings.size());
        assertNotNull(Server.bridgeSettings.get(0).keywordRemove);
        assertEquals(3, Server.bridgeSettings.get(0).keywordRemove.size());
        assertTrue(Server.bridgeSettings.get(0).keywordRemove.contains("opendata"));
        assertTrue(Server.bridgeSettings.get(0).keywordRemove.contains("foo"));
        assertTrue(Server.bridgeSettings.get(0).keywordRemove.contains("bar"));
    }

    @Test
    public void readEnvironmentVariables_multiple() {
        environmentVariables.set("BRIDGE_1_URL", "http://localhost:8081/csw");
        environmentVariables.set("BRIDGE_1_FIND_WMS_IN_INFO", "true");
        environmentVariables.set("BRIDGE_1_REMOVE_KEYWORDS", "opendata");
        environmentVariables.set("BRIDGE_1_FIX_MISSING_TEMPORAL", "1");
        environmentVariables.set("BRIDGE_1_TYPE_EXCLUDE", "application, tile,service");

        environmentVariables.set("BRIDGE_2_URL", "http://localhost:8082/csw");
        environmentVariables.set("BRIDGE_2_FIND_WMS_IN_INFO", "false");
        environmentVariables.set("BRIDGE_2_FILTER_OPENDATA", "true");
        environmentVariables.set("BRIDGE_2_SORT_RESULTS", "false");
        environmentVariables.set("BRIDGE_2_TYPE_INCLUDE", "dataset,series");
        environmentVariables.set("BRIDGE_2_IGNORE_SERIES_WITHOUT_DISTRIBUTION", "true");


        environmentVariables.set("BRIDGE_3_URL", "http://localhost:8083/csw");
        environmentVariables.set("BRIDGE_3_SORT_RESULTS", "1");
        Server.readEnvironmentVariables();

        BridgeSettings settings1 = Server.bridgeSettings.get(0);
        BridgeSettings settings2 = Server.bridgeSettings.get(1);
        BridgeSettings settings3 = Server.bridgeSettings.get(2);

        assertEquals(3, Server.bridgeSettings.size());
        assertEquals("http://localhost:8081/csw", settings1.cswURL);
        assertEquals("http://localhost:8082/csw", settings2.cswURL);
        assertEquals("http://localhost:8083/csw", settings3.cswURL);
        assertTrue(settings1.findWMSinInfo);
        assertFalse(settings2.findWMSinInfo);
        assertTrue(settings1.keywordRemove.contains("opendata"));
        assertTrue(settings1.addMissingTemporalStart);
        assertFalse(settings2.addMissingTemporalStart);
        assertTrue(settings2.filterOpenData);
        assertTrue(settings1.sortResults);
        assertFalse(settings2.sortResults);
        assertTrue(settings3.sortResults);
        assertFalse(settings1.ignoreSeriesWithoutDistribution);
        assertTrue(settings2.ignoreSeriesWithoutDistribution);
        assertFalse(settings1.ignoreSeriesWithoutDistribution);

        assertNotNull(settings1.typeExclude);
        assertEquals(3, settings1.typeExclude.length);
        assertEquals("[application, tile, service]", Arrays.toString(settings1.typeExclude));

        assertNotNull(settings2.typeInclude);
        assertEquals(2, settings2.typeInclude.length);
        assertEquals("[dataset, series]", Arrays.toString(settings2.typeInclude));
    }

    @Test
    public void handleRequest_invalidURL() throws IOException {
        new MockServerClient("127.0.0.1", mockServer.getPort())
                .when(
                        request()
                                .withMethod("POST")
                                .withHeader("X-CSW2DCATBridge", "getRecords")
                                .withPath("/csw"),
                        exactly(1))
                .respond(
                        response()
                                .withStatusCode(HttpStatusCode.OK_200.code())
                                .withHeaders(
                                        new Header("Content-Type", "text/xml; charset=utf-8"))
                                .withBody(IOUtils.toString(
                                        Objects.requireNonNull(getClass().getResourceAsStream("/GetRecordsResponse_mekun.xml")), StandardCharsets.UTF_8))
                );
        new MockServerClient("127.0.0.1", mockServer.getPort())
                .when(
                        request()
                                .withMethod("POST")
                                .withHeader("X-CSW2DCATBridge", "operatesOn")
                                .withPath("/csw"),
                        exactly(1))
                .respond(
                        response()
                                .withStatusCode(HttpStatusCode.OK_200.code())
                                .withHeaders(
                                        new Header("Content-Type", "text/xml; charset=utf-8"))
                                .withBody(IOUtils.toString(
                                        Objects.requireNonNull(getClass().getResourceAsStream("/GetRecordsResponse_operatesOn_mekun.xml")), StandardCharsets.UTF_8))
                );


        // mock HTTP request and response
        final HttpServerExchange exchange = Mockito.mock(HttpServerExchange.class);
        Mockito.when(exchange.getRequestPath()).thenReturn("/catalog.xml");
        final Sender sender = Mockito.mock(Sender.class);
        Mockito.when(exchange.getResponseSender()).thenReturn(sender);
        HeaderMap responseHeaders = new HeaderMap();
        Mockito.when(exchange.getResponseHeaders()).thenReturn(responseHeaders);

        final BridgeSettings bridgeSettings = new BridgeSettings();
        bridgeSettings.cswURL = "http://localhost:" + mockServer.getPort() + "/csw";

        Server.bridgeSettings.put(0, bridgeSettings);

        final Server server = new Server();

        // invoke method
        server.handleRequest(exchange);

        // check results
        assertEquals("application/rdf+xml", responseHeaders.getFirst(Headers.CONTENT_TYPE));
        assertEquals("*", responseHeaders.getFirst("Access-Control-Allow-Origin"));
    }
}
